// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// Red-Black Balanced Binary Tree Algorithm adapted from:
// Thomas Niemann <thomasn@epaperpress.com>
// See also:
// Thomas H. Cormen, et al
// Introduction to Algorithms
// (MIT Electrical Engineering and Computer Science)
// (C) 1990 MIT Press
//

type
  EStringTreeError = class (Exception);

  PStrRBTreeNode = ^TStrRBTreeNode;
  TStrRBTreeNode = record
    fIndex: Integer;
    fLeft: PStrRBTreeNode;
    fRight: PStrRBTreeNode;
    fParent: PStrRBTreeNode;
    fColor: Byte;
    fStr: string;
    fDColonPos: Integer;
  end;

  TStringTreeIterator = procedure (handle: Integer; const str: string);

  TStringTree = class
  protected
    fNodesUsed: Integer;
    fRoot: PStrRBTreeNode;
    fSentinel: PStrRBTreeNode;
    fPool: array of PStrRBTreeNode;
    fPoolUsed: Integer;
    fFreeList: PStrRBTreeNode;

    function GetStrings (handle: Integer): string;
    function GetDColonPos (handle: Integer): Integer;
    function AllocNode (): PStrRBTreeNode;
    procedure FreeNode (node: PStrRBTreeNode);
    function GetMaxHandle (): Integer;

  public
    constructor Create ();
    destructor Destroy (); override;

    procedure Clear ();

    function Append (const str: string): Integer; // insert string, return handle
    procedure Delete (handle: Integer);
    function Find (const str: string): Integer; // -1: no such string

    procedure Iterate (IterProc: TStringTreeIterator);

    function ValidHandle (handle: Integer): Boolean;

    property Count: Integer read fNodesUsed;
    property MaxHandle: Integer read GetMaxHandle;
    property Strings [handle: Integer]: string read GetStrings; default;
    property DColonPos [handle: Integer]: Integer read GetDColonPos;
  end;

var
  oneCharStrings: array [-257..-2] of string;


{ helper procedures }
procedure RotateLeft (tree: TStringTree; x: PStrRBTreeNode);
// rotate node x to left
var
  y: PStrRBTreeNode;
begin
  y := x.fRight;

  (* establish x.Right link *)
  x.fRight := y.fLeft;
  if y.fLeft <> tree.fSentinel then y.fLeft.fParent := x;

  (* establish y.Parent link *)
  if y <> tree.fSentinel then y.fParent := x.fParent;
  if x.fParent <> nil then
  begin
    if x = x.fParent.fLeft then x.fParent.fLeft := y else x.fParent.fRight := y;
  end
  else tree.fRoot := y;

  (* link x and y *)
  y.fLeft := x;
  if x <> tree.fSentinel then x.fParent := y;
end;

procedure RotateRight (tree: TStringTree; x: PStrRBTreeNode);
// rotate node x to right
var
  y: PStrRBTreeNode;
begin
  y := x.fLeft;

  (* establish x.Left link *)
  x.fLeft := y.fRight;
  if y.fRight <> tree.fSentinel then y.fRight.fParent := x;

  (* establish y.Parent link *)
  if y <> tree.fSentinel then y.fParent := x.fParent;
  if x.fParent <> nil then
  begin
    if x = x.fParent.fRight then x.fParent.fRight := y else x.fParent.fLeft := y;
  end
  else tree.fRoot := y;

  (* link x and y *)
  y.fRight := x;
  if x <> tree.fSentinel then x.fParent := y;
end;

procedure InsertFixup (tree: TStringTree; x: PStrRBTreeNode);
//  maintain red-black tree balance after inserting node x
var
  y: PStrRBTreeNode;
begin
  (* check red-black properties *)
  while (x <> tree.fRoot) and (x.fParent.fColor = RBT_RED) do
  begin
    (* we have a violation *)
    if x.fParent = x.fParent.fParent.fLeft then
    begin
      y := x.fParent.fParent.fRight;
      if y.fColor = RBT_RED then
      begin
        (* uncle is red *)
        x.fParent.fColor := RBT_BLACK;
        y.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        x := x.fParent.fParent;
      end
      else
      begin
        (* uncle is black *)
        if x = x.fParent.fRight then
        begin
          (* make x a fLeft child *)
          x := x.fParent;
          RotateLeft(tree, x);
        end;

        (* recolor and rotate *)
        x.fParent.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        RotateRight(tree, x.fParent.fParent);
      end;
    end
    else
    begin
      (* mirror image of above code *)
      y := x.fParent.fParent.fLeft;
      if y.fColor = RBT_RED then
      begin
        (* uncle is red *)
        x.fParent.fColor := RBT_BLACK;
        y.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        x := x.fParent.fParent;
      end
      else
      begin
        (* uncle is black *)
        if x = x.fParent.fLeft then
        begin
          x := x.fParent;
          RotateRight(tree, x);
        end;
        x.fParent.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        RotateLeft(tree, x.fParent.fParent);
      end;
    end;
  end;

  tree.fRoot.fColor := RBT_BLACK;
end;

procedure DeleteFixup (tree: TStringTree; x: PStrRBTreeNode);
// maintain red-black tree balance after deleting node x
var
  w: PStrRBTreeNode;
begin
  while (x <> tree.fRoot) and (x.fColor = RBT_BLACK) do
  begin
    if x = x.fParent.fLeft then
    begin
      w := x.fParent.fRight;
      if w.fColor = RBT_RED then
      begin
        w.fColor := RBT_BLACK;
        x.fParent.fColor := RBT_RED;
        RotateLeft(tree, x.fParent);
        w := x.fParent.fRight;
      end;
      if (w.fLeft.fColor = RBT_BLACK) and (w.fRight.fColor = RBT_BLACK) then
      begin
        w.fColor := RBT_RED;
        x := x.fParent;
      end
      else
      begin
        if w.fRight.fColor = RBT_BLACK then
        begin
          w.fLeft.fColor := RBT_BLACK;
          w.fColor := RBT_RED;
          RotateRight(tree, w);
          w := x.fParent.fRight;
        end;
        w.fColor := x.fParent.fColor;
        x.fParent.fColor := RBT_BLACK;
        w.fRight.fColor := RBT_BLACK;
        RotateLeft(tree, x.fParent);
        x := tree.fRoot;
      end
    end
    else
    begin
      w := x.fParent.fLeft;
      if w.fColor = RBT_RED then
      begin
        w.fColor := RBT_BLACK;
        x.fParent.fColor := RBT_RED;
        RotateRight(tree, x.fParent);
        w := x.fParent.fLeft;
      end;
      if (w.fRight.fColor = RBT_BLACK) and (w.fLeft.fColor = RBT_BLACK) then
      begin
        w.fColor := RBT_RED;
        x := x.fParent;
      end
      else
      begin
        if w.fLeft.fColor = RBT_BLACK then
        begin
          w.fRight.fColor := RBT_BLACK;
          w.fColor := RBT_RED;
          RotateLeft(tree, w);
          w := x.fParent.fLeft;
        end;
        w.fColor := x.fParent.fColor;
        x.fParent.fColor := RBT_BLACK;
        w.fLeft.fColor := RBT_BLACK;
        RotateRight(tree, x.fParent);
        x := tree.fRoot;
      end;
    end;
  end;

  x.fColor := RBT_BLACK;
end;


{ TStringTree }
constructor TStringTree.Create ();
var
  f: Integer;
begin
  for f := -257 to -2 do oneCharStrings[f] := chr(-(f)-2);
  SetLength(fPool, 0);
  Clear();
end;

destructor TStringTree.Destroy ();
begin
  Clear();
  Dispose(fPool[0]); // sentinel
  SetLength(fPool, 0);
end;

procedure TStringTree.Clear ();
var
  f: Integer;
begin
  f := fPoolUsed;
  while f > 0 do begin Dec(f); fPool[f].fStr := ''; Dispose(fPool[f]); end;
  SetLength(fPool, 0);

  fSentinel := nil;
  fPoolUsed := 0; fNodesUsed := 0; fFreeList := nil;
  fSentinel := AllocNode();
  fSentinel.fLeft := fSentinel; fSentinel.fRight := fSentinel;
  fSentinel.fParent := nil;
  fSentinel.fColor := RBT_BLACK;
  fRoot := fSentinel;
end;

function TStringTree.GetStrings (handle: Integer): string;
begin
  if (handle >= -257) and (handle <= -2) then
  begin
    result := oneCharStrings[handle];
    exit;
  end;
  if handle = 0 then begin result := ''; exit; end;
  if (handle < 1) or (handle >= fPoolUsed) or (fPool[handle].fColor = RBT_UNUSED) then
    raise EStringTreeError.Create('invalid string index');
  result := fPool[handle].fStr;
end;

function TStringTree.GetDColonPos (handle: Integer): Integer;
begin
  result := 0;
  if handle > 0 then
  begin
    if handle < fPoolUsed then result := fPool[handle].fDColonPos;
  end;
end;

procedure TStringTree.FreeNode (node: PStrRBTreeNode);
begin
  if (node = nil) or (node = fSentinel) then exit;
  ASSERT(node.fColor <> RBT_UNUSED);
  node.fLeft := fSentinel; node.fRight := fSentinel; node.fColor := RBT_UNUSED;
  node.fStr := '';
  node.fDColonPos := 0;
  node.fParent := fFreeList;
  fFreeList := node;
end;

function TStringTree.AllocNode (): PStrRBTreeNode;
var
  f: Integer;
begin
  if fFreeList <> nil then
  begin
    result := fFreeList;
    fFreeList := fFreeList.fParent;
  end
  else
  begin
    ASSERT(fPoolUsed <= Length(fPool));
    if fPoolUsed = Length(fPool) then
    begin
      SetLength(fPool, Length(fPool)+GROW_DELTA);
      for f := fPoolUsed to High(fPool) do fPool[f] := nil;
    end;
    New(result); fPool[fPoolUsed] := result;
    result.fDColonPos := 0; result.fStr := '';
    result.fLeft := fSentinel; result.fRight := fSentinel;
    result.fIndex := fPoolUsed; Inc(fPoolUsed);
  end;
  result.fParent := nil;
  result.fColor := RBT_RED;
end;

function TStringTree.Find (const str: string): Integer;
// -1: no such string
var
  r: Integer;
  node: PStrRBTreeNode;
begin
  if str = '' then begin result := 0; exit; end;
  if Length(str) = 1 then
  begin
    result := -(ord(str[1])+2);
    exit;
  end;

  result := -1; node := fRoot;
  while node <> fSentinel do
  begin
    r := CompareStr(str, node.fStr);
    if r = 0 then break;
    if r < 0 then node := node.fLeft else node := node.fRight;
  end;

  if node <> fSentinel then result := node.fIndex;
end;

function TStringTree.Append (const str: string): Integer;
// insert string, return handle
var
  current, parent, newNode: PStrRBTreeNode;
  r: Integer;
begin
  if str = '' then begin result := 0; exit; end;
  if Length(str) = 1 then
  begin
    result := -(ord(str[1])+2);
    exit;
  end;

  (* find where node belongs *)
  current := fRoot; parent := nil;
  while current <> fSentinel do
  begin
    r := CompareStr(str, current.fStr);
    if r = 0 then
    begin
      // duplicate found
      result := current.fIndex;
      exit;
    end;
    parent := current;
    if r < 0 then current := current.fLeft else current := current.fRight;
  end;

  (* setup new node *)
  newNode := AllocNode();
  newNode.fParent := parent;
  newNode.fLeft := fSentinel;
  newNode.fRight := fSentinel;
  newNode.fColor := RBT_RED;
  newNode.fStr := str;
  newNode.fDColonPos := Pos('::', str);

  (* insert node in tree *)
  if parent <> nil then
  begin
    if CompareStr(str, parent.fStr) < 0 then parent.fLeft := newNode
    else parent.fRight := newNode;
  end
  else fRoot := newNode;
  InsertFixup(self, newNode);

  Inc(fNodesUsed);
  result := newNode.fIndex;
end;

procedure TStringTree.Delete (handle: Integer);
var
  x, y, z: PStrRBTreeNode;
  color: Byte;
begin
  if handle = 0 then exit;
  if (handle >= -257) and (handle <= -2) then exit;

  if (handle < 1) or (handle >= fPoolUsed) or (fPool[handle].fColor = RBT_UNUSED) then
    raise EStringTreeError.Create('invalid string index');

  (* find node in tree *)
  z := fPool[handle];
  ASSERT(z <> fSentinel);

  if (z.fLeft = fSentinel) or (z.fRight = fSentinel) then
  begin
    (* y has a nil node as a child *)
    y := z;
  end
  else
  begin
    (* find tree successor with a sentinel node as a child *)
    y := z.fRight;
    while y.fLeft <> fSentinel do y := y.fLeft;
  end;

  (* x is y's only child *)
  if y.fLeft <> fSentinel then x := y.fLeft else x := y.fRight;

  (* remove y from the parent chain *)
  x.fParent := y.fParent;
  if y.fParent <> nil then
  begin
    if y = y.fParent.fLeft then y.fParent.fLeft := x else y.fParent.fRight := x;
  end
  else fRoot := x;

  color := y.fColor;

  if y <> z then
  begin
    (* swap y and z *)
    y.fLeft := z.fLeft;
    y.fRight := z.fRight;
    y.fParent := z.fParent;
    if z.fParent <> nil then
    begin
      if z.fParent.fLeft = z then z.fParent.fLeft := y else z.fParent.fRight := y;
    end
    else fRoot := y;

    y.fRight.fParent := y;
    y.fLeft.fParent := y;

    y.fColor := z.fColor;
  end;

  if color = RBT_BLACK then DeleteFixup(self, x);

  FreeNode(z);
  Dec(fNodesUsed);
end;

procedure TStringTree.Iterate (IterProc: TStringTreeIterator);
const
  STACK_SIZE = 50;
var
  stack: array [0..STACK_SIZE-1] of PStrRBTreeNode;
  sp: Integer;
  node: PStrRBTreeNode;
begin
  sp := 0; node := fRoot;
  repeat
    while node <> fSentinel do
    begin
      if sp >= STACK_SIZE then raise Exception.Create('out of stack in TStringTree.Iterate()');
      stack[sp] := node; Inc(sp);
      node := node.fLeft;
    end;
    if sp = 0 then break;

    Dec(sp); node := stack[sp];
    IterProc(node.fIndex, node.fStr);
    node := node.fRight;
  until false;
end;

function TStringTree.GetMaxHandle (): Integer;
begin
  result := fPoolUsed-1;
end;

function TStringTree.ValidHandle (handle: Integer): Boolean;
begin
  if (handle >= -257) and (handle <= -2) or (handle = 0) then begin result := true; exit; end;
  result := (handle > 0) and (handle < fPoolUsed) and (fPool[handle].fColor <> RBT_UNUSED);
end;
