// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// reading/writing .DZS format, v0.0.6
// code by Ketmar // Vampire Avalon
//
// RLE compression/decmpression code inspired by:
// z80.c, Philip Kendall, Darren Salt (part of libspectrum library)
//
// note: it ignores global MD5 for now %-(
//
{$DEFINE NTFS_CLEAR_COMPRESSION} // clear "compressed" attr on NTFS?
unit dzsLibU;

interface

uses
  SysUtils, Classes, xpackU, md5U, wpackU;


// machine codes
const
  dzsUnknown = 0;
  dzs16k = 1;
  dzs48k = 2;
  dzs128k = 3;
  dzsPlus2 = 4;
  dzsPentagon = 5;
  dzsPlus2a = 6;
  dzsPlus3 = 7;
  dzsScorpion = 8;
  dzsPlus3e = 9;

  dzsNoAY = 0;
  dzsAY = 1;
  dzsYM = 2;

  dzsPackNo = 0;
  dzsUnpacked = 0;
  dzsPackRLE = 1;
  dzsPackX = 2;
  dzsPackW = 3;


type
  EDZSError = class (Exception);

  TDZSZ80Reg = packed record
    case Integer of
      0: (l, h: Byte);
      1: (f, a: Byte);
      2: (r, i: Byte);
      3: (w: Word);
  end;

  TDZSZ80RegInfo = packed record
  //ACHTUNG: DO NOT CHANGE THE ORDER!!!
    bc, de, hl, af: packed array [0..1] of TDZSZ80Reg; // normal/alt
    ix, iy: TDZSZ80Reg; // ix/iy
    ir: TDZSZ80Reg;
    sp: TDZSZ80Reg;
    pc: TDZSZ80Reg;
  end;

  PDZSMemBlock = ^TDZSMemBlock;
  TDZSMemBlock = record
    next: PDZSMemBlock;
    page: Word; // or blockId for blocks
    data: packed array of Byte;
  end;

  PDZSZ80State = ^TDZSZ80State;
  TDZSZ80State = packed record
    model: Byte; // see below
    tStates: Integer; // t-states passed from the last interrupt
    regs: TDZSZ80RegInfo;
    im: Byte;
    iff1, iff2: Boolean;
    border: Byte;
    trdosIn: Boolean; // was tr-dos rom stamped in when the thing was saved?
    last7FFD: Byte;
    last1FFD: Byte;
    halted: Boolean; // was saved at HALT instruction
    lastWasEI: Boolean; // last executed instruction was EI
    ayType: Byte;
    lastFFFD: Byte; // ay/ym
    psg: packed array [0..15] of Byte; // ay/ym registers
    memory: PDZSMemBlock;  // ,    16 
    blocks: PDZSMemBlock;  // 
    unkBlocks: PDZSMemBlock;  //  
  end;

  TDZSBlockInfo = record
    blockType: Byte;
    compression: Byte;
    packDataSize :Integer;
    dataSize: Integer;
    hasMD5: Boolean;
    md5: TMD5Digest;
    curMD5: TMD5Context;
  end;

  TDZSCustomBlock = procedure (data: Pointer; dataSize: Integer);
    // custom block (13) processor


function DZSNewStateRec (): PDZSZ80State;
procedure DZSFreeStateRec (var state: PDZSZ80State);
procedure DZSClearStateRec (state: PDZSZ80State);
function DZSPeek (state: PDZSZ80State; page: Byte; addr: Word): Byte;
procedure DZSPoke (state: PDZSZ80State; page: Byte; addr: Word; value: Byte);
function DZSNewPage (state: PDZSZ80State; page: Byte): PDZSMemBlock;
  // adds new page even if there alrady was such one!
function DZSNewPageWithData (state: PDZSZ80State; page: Byte;
  const data; dataSize: Integer): PDZSMemBlock;
  // adds new page even if there alrady was such one!
function DZSNewAddonBlock (state: PDZSZ80State; blockId: Word;
  const data; dataSize: Integer): PDZSMemBlock;
  // adds new block even if there alrady was such one!
function DZSFindPage (state: PDZSZ80State; page: Byte): PDZSMemBlock;
  // latest one
function DZSFindAddonBlock (state: PDZSZ80State; id: Word): PDZSMemBlock;
  // latest one
function DZSNewUnkBlock (state: PDZSZ80State; blockId: Byte; const data; dataSize: Integer): PDZSMemBlock;
function DZSFindUnkBlock (state: PDZSZ80State; id: Byte): PDZSMemBlock;

{ snapshot i/o }
procedure DZSSaveToStream (st: TStream; const state: PDZSZ80State; compression: Byte = 3);
procedure DZSSaveToFile (const fName: string; const state: PDZSZ80State; compression: Byte = 3);
procedure DZSLoadFromStream (st: TStream; var state: PDZSZ80State; keepUnknownBlocks: Boolean = false);
procedure DZSLoadFromFile (const fName: string; var state: PDZSZ80State; keepUnknownBlocks: Boolean = false);

{ .Z80 RLE compression & decompression }
function Z80RLEPackBuffer (const bufIn; inSize: Integer; out bufOut; outSize: Integer): Integer;
// -1: bufOut filled, but there's some data to compress left
// any positive value (include 0): no of compressed bytes (i.e. bytes, used in bufOut)
function Z80RLEUnpackBuffer (const bufIn; inSize: Integer; out bufOut; outSize: Integer): Integer;
// -1: bufOut filled, but there's some uncompressed data
// any positive value (include 0): # of decompressed bytes (i.e. bytes, used in bufOut)

{ additional snapshot formats }
procedure DZSLoadSNAFromStream (st: TStream; var state: PDZSZ80State);
procedure DZSLoadZ80FromStream (st: TStream; var state: PDZSZ80State);
procedure DZSLoadSNA (const fName: string; var state: PDZSZ80State);
procedure DZSLoadZ80 (const fName: string; var state: PDZSZ80State);
procedure DZSSaveSNAToStream (st: TStream; var state: PDZSZ80State);
procedure DZSSaveZ80ToStream (st: TStream; const state: PDZSZ80State; compression: Byte = 1);
procedure DZSSaveSNA (const fName: string; var state: PDZSZ80State);
procedure DZSSaveZ80 (const fName: string; const state: PDZSZ80State; compression: Byte = 1);


procedure NTFSClearCompression (fh: Cardinal);
procedure NTFSFileClearCompression (const fileName: string);

function CreateFileStream (const fName: string): TFileStream;
  // fmCreate/fmOpenReadWrite
  // throws an exception on error


implementation

{$IFDEF NTFS_CLEAR_COMPRESSION}
uses
  Windows;
{$ENDIF}


procedure NTFSClearCompression (fh: Cardinal);
{$IFDEF NTFS_CLEAR_COMPRESSION}
var
  bytesReturned: Cardinal;
  buffer: Short;
{$ENDIF}
begin
{$IFDEF NTFS_CLEAR_COMPRESSION}
  buffer := 0;
  DeviceIoControl(fh, $0009C040, @buffer, SizeOf(Short), nil, 0, bytesReturned, nil);
{$ENDIF}
end;

procedure NTFSFileClearCompression (const fileName: string);
{$IFDEF NTFS_CLEAR_COMPRESSION}
var
  fh: Integer;
{$ENDIF}
begin
{$IFDEF NTFS_CLEAR_COMPRESSION}
  fh := FileOpen(fileName, fmOpenReadWrite or fmShareExclusive);
  if fh > -1 then
  begin
    NTFSClearCompression(fh);
    FileClose(fh);
  end;
{$ENDIF}
end;


function DZSNewStateRec (): PDZSZ80State;
begin
  New(result);
  FillChar(result^, SizeOf(TDZSZ80State), 0);
end;

procedure DZSFreeStateRec (var state: PDZSZ80State);
begin
  if state <> nil then begin DZSClearStateRec(state); Dispose(state); end;
  state := nil;
end;

procedure DZSClearStateRec (state: PDZSZ80State);
var
  blk, nblk: PDZSMemBlock;
begin
  blk := state.memory;
  while blk <> nil do begin nblk := blk.next; blk.data := nil; Dispose(blk); blk := nblk; end;
  blk := state.blocks;
  while blk <> nil do begin nblk := blk.next; blk.data := nil; Dispose(blk); blk := nblk; end;
  blk := state.unkBlocks;
  while blk <> nil do begin nblk := blk.next; blk.data := nil; Dispose(blk); blk := nblk; end;
  FillChar(state^, SizeOf(TDZSZ80State), 0);
end;

function DZSNewPageWithData (state: PDZSZ80State; page: Byte;
  const data; dataSize: Integer): PDZSMemBlock;
var
  blk, nblk: PDZSMemBlock;
begin
  New(nblk);
  nblk.next := nil; nblk.page := page;
  try
    if dataSize > 16384 then dataSize := 16384;
    SetLength(nblk.data, 16384);
    FillChar(nblk.data[0], 16384, 0);
    if dataSize > 0 then Move(data, nblk.data[0], dataSize);
    blk := state.memory;
    if blk <> nil then
    begin
      while blk.next <> nil do blk := blk.next;
      blk.next := nblk;
    end
    else state.memory := nblk;
    result := nblk;
  except
    nblk.data := nil; Dispose(nblk);
    raise;
  end;
end;

function DZSNewPage (state: PDZSZ80State; page: Byte): PDZSMemBlock;
begin
  result := DZSNewPageWithData(state, page, (nil)^, 0);
end;

function DZSNewAddonBlock (state: PDZSZ80State; blockId: Word; const data; dataSize: Integer): PDZSMemBlock;
var
  blk, nblk: PDZSMemBlock;
begin
  //result := nil;
  //if dataSize = 0 then exit; //!
  if (dataSize < 0) or (dataSize > 1024*1024) then raise EDZSError.Create('invalid block data size');
  New(nblk); nblk.next := nil; nblk.page := blockId;
  try
    SetLength(nblk.data, dataSize);
    if dataSize > 0 then Move(data, nblk.data[0], dataSize);
    blk := state.blocks;
    if blk <> nil then
    begin
      while blk.next <> nil do blk := blk.next;
      blk.next := nblk;
    end
    else state.blocks := nblk;
    result := nblk;
  except
    nblk.data := nil; Dispose(nblk);
    raise;
  end;
end;

function DZSFindPage (state: PDZSZ80State; page: Byte): PDZSMemBlock;
var
  blk: PDZSMemBlock;
begin
  result := nil;
  blk := state.memory;
  while blk <> nil do
  begin
    if blk.page = page then result := blk;
    blk := blk.next;
  end;
end;

function DZSFindAddonBlock (state: PDZSZ80State; id: Word): PDZSMemBlock;
var
  blk: PDZSMemBlock;
begin
  result := nil;
  blk := state.blocks;
  while blk <> nil do
  begin
    if blk.page = id then result := blk;
    blk := blk.next;
  end;
end;

function DZSNewUnkBlock (state: PDZSZ80State; blockId: Byte; const data; dataSize: Integer): PDZSMemBlock;
var
  blk, nblk: PDZSMemBlock;
begin
  //result := nil;
  //if dataSize = 0 then exit; //!
  if (dataSize < 0) or (dataSize > 1024*1024) then raise EDZSError.Create('invalid block data size');
  New(nblk); nblk.next := nil; nblk.page := blockId;
  try
    SetLength(nblk.data, dataSize);
    if dataSize > 0 then Move(data, nblk.data[0], dataSize);
    blk := state.unkBlocks;
    if blk <> nil then
    begin
      while blk.next <> nil do blk := blk.next;
      blk.next := nblk;
    end
    else state.unkBlocks := nblk;
    result := nblk;
  except
    nblk.data := nil; Dispose(nblk);
    raise;
  end;
end;

function DZSFindUnkBlock (state: PDZSZ80State; id: Byte): PDZSMemBlock;
var
  blk: PDZSMemBlock;
begin
  result := nil;
  blk := state.unkBlocks;
  while blk <> nil do
  begin
    if blk.page = id then result := blk;
    blk := blk.next;
  end;
end;

function DZSPeek (state: PDZSZ80State; page: Byte; addr: Word): Byte;
var
  blk: PDZSMemBlock;
begin
  blk := DZSFindPage(state, page);
  if blk = nil then result := 0
  else if addr > High(blk.data) then result := 0
  else result := blk.data[addr];
end;

procedure DZSPoke (state: PDZSZ80State; page: Byte; addr: Word; value: Byte);
var
  blk: PDZSMemBlock;
  f: Integer;
begin
  if addr > 16383 then exit;
  blk := DZSFindPage(state, page);
  if blk = nil then blk := DZSNewPage(state, page);
  f := Length(blk.data);
  if f < 16384 then
  begin
    SetLength(blk.data, 16384);
    while f <= High(blk.data) do begin blk.data[f] := 0; Inc(f); end;
  end;
  blk.data[addr] := value;
end;



function Z80RLEPackBuffer (const bufIn; inSize: Integer; out bufOut; outSize: Integer): Integer;
// -1: bufOut filled, but there's some data to compress left
// any positive value (include 0): no of compressed bytes (i.e. bytes, used in bufOut)
var
  inP, outP: PChar;
  cnt: Integer;
  filler: Char;
  lastED: Boolean;
begin
  inP := @bufIn; outP := @bufOut; lastED := false; result := 0;
  while (result < outSize) and (inSize > 0) do
  begin
    if inSize = 1 then
    begin
      // the last byte. just copy it & go out
      // no checks, 'cause there's at least one free byte in <outBuf> anyway
      outP^ := inP^; Inc(result); Dec(inSize); break;
    end;
    // here we have at least two bytes in <inBuf>
    // check if we're pointing to a run of identical bytes & the last output one wasn't a single $ED
    if not lastED and (inP[0] = inP[1]) then
    begin
      filler := inP^; cnt := 2;
      Inc(inP, 2); Dec(inSize, 2);
      // find the length of the run (but not more than 255 bytes!)
      while (cnt < 255) and (inSize > 0) and (inP^ = filler) do
      begin
        Inc(inP); Dec(inSize); Inc(cnt);
      end;
      // Output this in compressed form if it's of length 5 or longer, or if it's a run of $ED
      if (cnt >= 5) or (filler = #$ED) then
      begin
        if result+4 > outSize then begin result := -1; break; end; // no room in <outBuf>
        outP[0] := #$ED; outP[1] := #$ED; outP[2] := chr(cnt); outP[3] := filler;
        Inc(outP, 4); Inc(result, 4);
      end
      else
      begin
        // just output the bytes
        if result+cnt > outSize then begin result := -1; break; end; // no room in <outBuf>
        while cnt > 0 do
        begin
          outP^ := filler; Inc(outP); Inc(result); Dec(cnt);
        end;
      end
    end
    else
    begin
      // not a repeated character (or can't insert $ED here), so just output the byte
      lastED := (inP^ = #$ED);
      outP^ := inP^;
      Inc(inP); Dec(inSize); Inc(outP); Inc(result);
    end;
  end;

  // check if everything was compressed
  if inSize <> 0 then result := -1; // alas. out of <outBuf>
end;

function Z80RLEUnpackBuffer (const bufIn; inSize: Integer; out bufOut; outSize: Integer): Integer;
// -1: bufOut filled, but there's some uncompressed data
// any positive value (include 0): # of decompressed bytes (i.e. bytes, used in bufOut)
var
  inP, outP: PChar;
  cnt: Integer;
  filler: Char;
begin
  inP := @bufIn; outP := @bufOut; result := 0;
  while (result < outSize) and (inSize > 0) do
  begin
    if inSize = 1 then
    begin
      // the last byte. just copy it & go out
      // no checks, 'cause there's at least one free byte in <outBuf> anyway
      outP^ := inP^; Inc(result); Dec(inSize); break;
    end;
    // here we have at least two bytes in <inBuf>
    if (inP[0] = #$ED) and (inP[1] = #$ED) then
    begin
      // run (filler)
      Inc(inP, 2); Dec(inSize, 2);
      if inSize < 2 then raise Exception.Create('Z80: invalid compressed block %-(');
      cnt := ord(inP[0]); filler := inP[1];
      Inc(inP, 2); Dec(inSize, 2);
      while (cnt > 0) and (result < outSize) do
      begin
        outP^ := filler; Inc(outP); Inc(result); Dec(cnt);
      end;
      if (result = outSize) and (cnt > 0) then begin result := -1; break; end; // out of <outBuf>
    end
    else
    begin
      // just copy
      outP^ := inP^;
      Inc(inP); Dec(inSize); // no checks, 'cause at least two bytes is always here
      Inc(outP); Inc(result); // and there was at least one free byte in <outBuf>
    end;
  end;

  // check if everything was decompressed
  if inSize <> 0 then result := -1; // alas. out of <outBuf>
end;


procedure DZSWriteBlockFlagsAndSize (st: TStream; pdsize, udsize: Integer; compression: Byte;
  md5: Pointer);
// md5=nil: no MD5
// pdsize: packed data size
// udsize: unpacked data size
var
  b: Byte;
  psz, usz: Byte;
begin
  ASSERT(compression <= 7);
  if compression = 0 then pdsize := 0;
  ASSERT(pdsize >= 0);
  ASSERT(udsize >= 0);
  if pdsize <= $FF then psz := 0
  else if pdsize <= $FFFF then psz := 1
  else if pdsize <= $FFFFFF then psz := 2
  else psz := 3;
  if udsize <= $FF then usz := 0
  else if udsize <= $FFFF then usz := 1
  else if udsize <= $FFFFFF then usz := 2
  else usz := 3;
  b := compression or (psz shl 3) or (usz shl 5);
  if md5 <> nil then b := b or $80;
  st.WriteBuffer(b, 1);
  if md5 <> nil then st.WriteBuffer(md5^, 16);
  //FIXME: ENDIANESS! %-)
  st.WriteBuffer(udsize, usz+1);
  if compression <> 0 then st.WriteBuffer(pdsize, psz+1);
end;

procedure DZSWriteBlock (st: TStream; compression: Byte; blockType: Byte;
  const data; dataSize: Integer; writeMD5: Boolean);
var
  obuf: packed array of Byte;
  b: Byte;
  packSz: Integer;
  md5: TMD5Digest;
  md5ctx: TMD5Context;
  md5p: Pointer;
  ms: TMemoryStream;
begin
  ASSERT(compression <= 3);
  obuf := nil;

  //if dataSize < 1 then compression := dzsPackNo;

  //WriteLn('type: ', blockType);
  //WriteLn('dataSize: ', dataSize);
  if dataSize > 0 then
  begin
    SetLength(obuf, dataSize-1); packSz := -1;

    if compression = dzsPackW then
    begin
      packSz := WDXPack(data, dataSize, obuf[0], Length(obuf));
      //!WriteLn('x: ', packSz);
      if packSz < 0 then compression := dzsPackX;
        // try X packer, althru this is almost useless %-)
    end;

    if compression = dzsPackX then
    begin
      packSz := XPackBuffer(data, dataSize, obuf[0], Length(obuf));
      //!WriteLn('x: ', packSz);
      if packSz < 0 then compression := dzsPackRLE;
        // try RLE packer, althru this is almost useless %-)
    end;

    if compression = dzsPackRLE then
    begin
      packSz := Z80RLEPackBuffer(data, dataSize, obuf[0], Length(obuf));
      //if packSz < 0 then compression := dzsPackNo;
        // fall to uncompressed data
    end;

    if packSz < 0 then
    begin
      compression := dzsPackNo;
      packSz := -1;
    end;
  end
  else begin packSz := -1; compression := dzsPackNo; end;

  if writeMD5 then
  begin
    md5p := @md5;
    MD5Init(md5ctx);
    MD5Update(md5ctx, blockType, 1);
    ms := TMemoryStream.Create();
    try
      DZSWriteBlockFlagsAndSize(ms, packSz, dataSize, compression, md5p);
      ms.Position := 0;
      ms.ReadBuffer(b, 1); MD5Update(md5ctx, b, 1);
      ms.Position := 17; // skip MD5
      while ms.Position < ms.Size do
      begin
        ms.ReadBuffer(b, 1);
        MD5Update(md5ctx, b, 1);
      end;
    finally
      ms.Free();
    end;
    if dataSize > 0 then MD5Update(md5ctx, data, dataSize);
    md5 := MD5Final(md5ctx);
  end
  else md5p := nil;

  //!WriteLn('packSz: ', packSz);
  st.WriteBuffer(blockType, 1);
  DZSWriteBlockFlagsAndSize(st, packSz, dataSize, compression, md5p);
  if packSz < 0 then
  begin
    if dataSize > 0 then st.WriteBuffer(data, dataSize);
  end
  else if packSz > 0 then st.WriteBuffer(obuf[0], packSz);
end;

procedure DZSWritePageBlock (st: TStream; const state: PDZSZ80State;
  blk: PDZSMemBlock; compression: Byte; isBlock: Boolean);
var
  nbuf: packed array of Byte;
  b: Byte;
begin
  ASSERT(compression <= 3);
  if (blk = nil) or (blk.data = nil) then exit; // no such page!
  if not isBlock and (blk.page > 255) then raise EDZSError.Create('invalid page number');

  if isBlock then
  begin
    SetLength(nbuf, Length(blk.data)+2);
    FillChar(nbuf[0], Length(nbuf), 0);
    nbuf[0] := blk.page and $FF;
    nbuf[1] := (blk.page shr 8) and $FF;
    Move(blk.data[0], nbuf[2], Length(blk.data));
    // block type: "add. block" block ()
    b := 11;
  end
  else
  begin
    SetLength(nbuf, 16385);
    FillChar(nbuf[0], Length(nbuf), 0);
    nbuf[0] := blk.page;
    if Length(blk.data) < 16384 then Move(blk.data[0], nbuf[1], Length(blk.data))
    else Move(blk.data[0], nbuf[1], 16384);
    // block type: page block
    b := 1;
  end;

  DZSWriteBlock(st, compression, b, nbuf[0], Length(nbuf), true);
end;

procedure DZSReadBlockInfo (st: TStream; out info: TDZSBlockInfo);
//  blockType: Byte;
//  compression: Byte;
//  packDataSize: Integer;
//  dataSize: Integer;
//  hasMD5: Boolean;
//  md5: TMD5Digest;
//  curMD5: TMD5Context;
var
  b, bflg, ssz: Byte;
begin
  st.ReadBuffer(info.blockType, 1); // block type
  st.ReadBuffer(bflg, 1); // flagz

  info.compression := (bflg and $07);
  info.hasMD5 := (bflg and $80) <> 0;
  if info.hasMD5 then st.ReadBuffer(info.md5, SizeOf(info.md5))
  else FillChar(info.md5, SizeOf(info.md5), 0);

  MD5Init(info.curMD5);
  MD5Update(info.curMD5, info.blockType, 1);
  MD5Update(info.curMD5, bflg, 1);

  // read unpacked size
  ssz := (bflg shr 5) and $03;
  st.ReadBuffer(b, 1); info.dataSize := b;
  MD5Update(info.curMD5, b, 1);
  if ssz > 0 then
  begin
    st.ReadBuffer(b, 1); info.dataSize := info.dataSize+b*256;
    MD5Update(info.curMD5, b, 1);
  end;
  if ssz > 1 then
  begin
    st.ReadBuffer(b, 1); info.dataSize := info.dataSize+b*256*256;
    MD5Update(info.curMD5, b, 1);
  end;
  if ssz > 2 then
  begin
    st.ReadBuffer(b, 1); info.dataSize := info.dataSize+b*256*256*256;
    MD5Update(info.curMD5, b, 1);
  end;
  if info.dataSize < 0 then raise EDZSError.Create('invalid block size');

  // read packed size (if any)
  if info.compression <> 0 then
  begin
    ssz := (bflg shr 3) and $03;
    st.ReadBuffer(b, 1); info.packDataSize := b;
    MD5Update(info.curMD5, b, 1);
    if ssz > 0 then
    begin
      st.ReadBuffer(b, 1); info.packDataSize := info.packDataSize+b*256;
      MD5Update(info.curMD5, b, 1);
    end;
    if ssz > 1 then
    begin
      st.ReadBuffer(b, 1); info.packDataSize := info.packDataSize+b*256*256;
      MD5Update(info.curMD5, b, 1);
    end;
    if ssz > 2 then
    begin
      st.ReadBuffer(b, 1); info.packDataSize := info.packDataSize+b*256*256*256;
      MD5Update(info.curMD5, b, 1);
    end;
    if info.packDataSize < 1 then raise EDZSError.Create('invalid packed block size');
  end
  else info.packDataSize := -1;
end;

procedure DZSSkipBlockData (st: TStream; const info: TDZSBlockInfo);
begin
  if info.compression = 0 then st.Position := st.Position+info.dataSize
  else st.Position := st.Position+info.packDataSize;
end;

procedure DZSReadBlockData (st: TStream; var info: TDZSBlockInfo; out data);
var
  pbuf: packed array of Byte;
  md5c: TMD5Digest;
  rd: Integer;
begin
  pbuf := nil;
  if info.compression = dzsPackNo then
  begin
    if info.dataSize > 0 then st.ReadBuffer(data, info.dataSize);
  end
  else if info.compression = dzsPackW then
  begin
    SetLength(pbuf, info.packDataSize);
    st.ReadBuffer(pbuf[0], info.packDataSize);
    rd := WDXUnpack(pbuf[0], info.packDataSize, data, info.dataSize);
    //!WriteLn(rd, ' | ', info.dataSize, ' | ', info.packDataSize);
    if rd = -1 then raise EDZSError.Create('invalid w-packed block');
  end
  else if info.compression = dzsPackX then
  begin
    SetLength(pbuf, info.packDataSize);
    st.ReadBuffer(pbuf[0], info.packDataSize);
    rd := XUnpackBuffer(pbuf[0], info.packDataSize, data, info.dataSize);
    //!WriteLn(rd, ' | ', info.dataSize, ' | ', info.packDataSize);
    if rd <> info.dataSize then raise EDZSError.Create('invalid x-packed block');
  end
  else if info.compression = dzsPackRLE then
  begin
    SetLength(pbuf, info.packDataSize);
    st.ReadBuffer(pbuf[0], info.packDataSize);
    rd := Z80RLEUnpackBuffer(pbuf[0], info.packDataSize, data, info.dataSize);
    if rd <> info.dataSize then raise EDZSError.Create('invalid rle-packed block');
  end
  else raise EDZSError.Create('invalid packer type');

  if info.dataSize > 0 then MD5Update(info.curMD5, data, info.dataSize);
  if info.hasMD5 then
  begin
    // check MD5
    md5c := MD5Final(info.curMD5);
    if not MD5Equal(info.md5, md5c) then raise EDZSError.Create('invalid block MD5');
  end;
end;

procedure DZSSaveToStream (st: TStream; const state: PDZSZ80State; compression: Byte = 3);
var
  sign: packed array [0..3] of Char;
  b: Byte;
  blk: PDZSMemBlock;
  ms: TMemoryStream;
  data: packed array of Byte;
  //md5: TMD5Digest;
begin
  //ASSERT((compression >= 0) and (compression <= 2));
  ASSERT(compression <= 3);
  data := nil;
  if (compression > dzsPackX) then
  begin
    if st is THandleStream then NTFSClearCompression(THandleStream(st).Handle);
  end;

  // signature & version
  sign := 'DZS'#3;
  st.WriteBuffer(sign[0], 4);

  // save block "type 0" -- machine state information
  ms := TMemoryStream.Create();
  try
    ms.WriteBuffer(state.model, 1);
    ms.WriteBuffer(state.regs, SizeOf(state.regs)); // 26 bytes

    b := (state.im and $03);
    if state.iff1 then b := b or $04;
    if state.iff2 then b := b or $08;
    if state.trdosIn then b := b or $10;
    if state.ayType = dzsAY then b := b or $20
    else if state.ayType = dzsYM then b := b or $40;
    if state.halted then b := b or $80;
    ms.WriteBuffer(b, 1);

    b := (state.border and $07);
    if state.lastWasEI then b := b or $80;
    ms.WriteBuffer(b, 1);

    if state.model <= dzs48k then b := $10 else b := state.last7FFD;
    ms.WriteBuffer(b, 1);
    if state.model < dzsPentagon then b := $FF else b := state.last1FFD;
    ms.WriteBuffer(b, 1);

    // ay/fm state
    ms.WriteBuffer(state.lastFFFD, 1);
    ms.WriteBuffer(state.psg[0], 16);

    // t-states
    b := (state.tStates and $FF); ms.WriteBuffer(b, 1);
    b := ((state.tStates shr 8) and $FF); ms.WriteBuffer(b, 1);
    b := ((state.tStates shr 16) and $FF); ms.WriteBuffer(b, 1);

    ms.Position := 0;
    SetLength(data, ms.Size); ms.ReadBuffer(data[0], Length(data));
  finally
    ms.Free();
  end;
  DZSWriteBlock(st, compression, 0, data[0], Length(data), true);
  data := nil;

  if state.model = dzs16k then DZSWritePageBlock(st, state, DZSFindPage(state, 5), compression, false)
  else if state.model = dzs48k then
  begin
    DZSWritePageBlock(st, state, DZSFindPage(state, 5), compression, false);
    DZSWritePageBlock(st, state, DZSFindPage(state, 2), compression, false);
    DZSWritePageBlock(st, state, DZSFindPage(state, 0), compression, false);
  end
  else
  begin
    blk := state.memory;
    while blk <> nil do
    begin
      DZSWritePageBlock(st, state, blk, compression, false);
      blk := blk.next;
    end;
  end;
  // write add. blocks ()
  blk := state.blocks;
  while blk <> nil do
  begin
    DZSWritePageBlock(st, state, blk, compression, true);
    blk := blk.next;
  end;
  // write other blocks
  blk := state.unkBlocks;
  while blk <> nil do
  begin
    DZSWriteBlock(st, compression, blk.page, blk.data[0], Length(blk.data), true);
    blk := blk.next;
  end;
end;

function CreateFileStream (const fName: string): TFileStream;
begin
  try
    result := TFileStream.Create(fName, fmCreate);
  except
    result := TFileStream.Create(fName, fmOpenReadWrite or fmShareDenyWrite);
  end;
end;

procedure DZSSaveToFile (const fName: string; const state: PDZSZ80State; compression: Byte = 3);
var
  st: TFileStream;
begin
  st := CreateFileStream(fName);//TFileStream.Create(fName, fmCreate);
  {.$IFDEF NTFS_CLEAR_COMPRESSION}
  //if compression >= dzsPackX then NTFSClearCompression(st.Handle);
  {.$ENDIF}
  try
    try
      DZSSaveToStream(st, state, compression);
    finally
      st.Free();
    end;
  except
    SysUtils.DeleteFile(fName);
    raise;
  end;
end;

procedure DZSLoadFromStream (st: TStream; var state: PDZSZ80State; keepUnknownBlocks: Boolean = false);
var
  sign: packed array [0..2] of Char;
  md5: TMD5Digest;
  info: TDZSBlockInfo;
  buf: packed array of Byte;
  bufPos: Integer;
  b: Byte;
  f: Integer;

  procedure ReadBuf (out b; size: Integer);
  begin
    if bufPos >= info.dataSize then raise EDZSError.Create('block '+IntToStr(info.blockType)+' is too small');
    Move(buf[bufPos], b, size); Inc(bufPos, size);
  end;

begin
  DZSClearStateRec(state); buf := nil;
  try
    st.ReadBuffer(sign, 3);
    if sign <> 'DZS' then raise EDZSError.Create('invalid signature');
    st.ReadBuffer(b, 1);
    if (b and $3F) <> 3 then raise EDZSError.Create('invalid version');
    if (b and $80) <> 0 then st.ReadBuffer(md5, SizeOf(md5)); // and ignore %-)

    // blocks
    buf := nil;
    while st.Position < st.Size do
    begin
      bufPos := 0;
      DZSReadBlockInfo(st, info);
      //*WriteLn('DZS block: ', info.blockType);
      if info.dataSize < 1 then begin DZSSkipBlockData(st, info); continue; end;

      case info.blockType of
        0: // machine state block
          begin
            if info.dataSize <> 51 then raise EDZSError.Create('invalid machine info block size');
            if Length(buf) < info.dataSize then SetLength(buf, info.dataSize);
            DZSReadBlockData(st, info, buf[0]);

            ReadBuf(state.model, 1);
            if (state.model < dzs16k) or (state.model > dzsPlus3e) then
              raise EDZSError.Create('invalid model');
            ReadBuf(state.regs, SizeOf(state.regs));
            ReadBuf(b, 1);
            state.im := (b and $03);
            if state.im = 3 then raise EDZSError.Create('invalid IM mode');
            state.iff1 := (b and $04) <> 0;
            state.iff2 := (b and $08) <> 0;
            state.trdosIn := (b and $10) <> 0;
            b := (b shr 5) and $03;
            if b = 3 then raise EDZSError.Create('invalid AY/FM model');
            state.ayType := b;
            state.halted := (b and $80) <> 0;

            ReadBuf(b, 1);
            state.border := (b and $07);
            state.lastWasEI := (b and $80) <> 0;

            ReadBuf(b, 1);
            if state.model > dzs48k then state.last7FFD := b else state.last7FFD := $10;

            ReadBuf(b, 1);
            if state.model >= dzsPentagon then state.last1FFD := b else state.last1FFD := $FF;

            // ay/fm state
            ReadBuf(state.lastFFFD, 1);
            ReadBuf(state.psg[0], 16);
            if state.ayType = dzsNoAY then
            begin
              state.lastFFFD := 0;
              FillChar(state.psg[0], 16, 0);
            end;

            // t-states
            ReadBuf(b, 1); state.tStates := b;
            ReadBuf(b, 1); state.tStates := state.tStates+b*256;
            ReadBuf(b, 1); state.tStates := state.tStates+b*65536;
          end;
        1: // RAM page block
          begin
            if info.dataSize > 16385 then raise EDZSError.Create('invalid RAM page block size');
            if Length(buf) < info.dataSize then SetLength(buf, info.dataSize);
            DZSReadBlockData(st, info, buf[0]);

            //ReadBuf(b, 1); // page no
            //if b < 8 then ReadBuf(state.mem[b], 16384);
            DZSNewPage(state, buf[0]);
            for f := 1 to info.dataSize-1 do
            begin
              DZSPoke(state, buf[0], f-1, buf[f]);
            end;
          end;
        11: // add. block block
          begin
            if (info.dataSize < 3) or (info.dataSize > 1024*1024) then raise EDZSError.Create('invalid "add. block" block size');
            if Length(buf) < info.dataSize then SetLength(buf, info.dataSize);
            DZSReadBlockData(st, info, buf[0]);
            DZSNewAddonBlock(state, buf[0]+256*buf[1], buf[2], info.dataSize-2);
          end;
        else
          begin
            if keepUnknownBlocks then
            begin
              if (info.dataSize < 0) or (info.dataSize > 1024*1024) then raise EDZSError.Create('invalid unknown block size');
              if Length(buf) < info.dataSize then SetLength(buf, info.dataSize);
              DZSReadBlockData(st, info, buf[0]);
              DZSNewUnkBlock(state, info.blockType, buf[0], info.dataSize);
            end
            else DZSSkipBlockData(st, info);
          end;
      end;
    end;
  except
    DZSClearStateRec(state);
    raise;
  end;
end;

procedure DZSLoadFromFile (const fName: string; var state: PDZSZ80State; keepUnknownBlocks: Boolean = false);
var
  st: TStream;
begin
  st := TFileStream.Create(fName, fmOpenRead or fmShareDenyWrite);
  try
    try
      DZSLoadFromStream(st, state, keepUnknownBlocks);
    finally
      st.Free();
    end;
  except
    raise;
  end;
end;

///////////////////////////////////////////
procedure DZSLoadSNA (const fName: string; var state: PDZSZ80State);
var
  st: TStream;
begin
  st := TFileStream.Create(fName, fmOpenRead or fmShareDenyWrite);
  try
    DZSLoadSNAFromStream(st, state);
  finally
    st.Free();
  end;
end;

procedure DZSLoadZ80 (const fName: string; var state: PDZSZ80State);
var
  st: TStream;
begin
  st := TFileStream.Create(fName, fmOpenRead or fmShareDenyWrite);
  try
    DZSLoadZ80FromStream(st, state);
  finally
    st.Free();
  end;
end;

procedure DZSLoadSNAFromStream (st: TStream; var state: PDZSZ80State);
const
  bnk: array [0..3] of Byte = (0, 5, 2, 0);
var
  f: Integer;
  b: Byte;
  tmp: packed array [0..16383] of Byte;
begin
  DZSClearStateRec(state);
  try
    if (st.Size <> 49179) and (st.Size <> 131103) then raise Exception.Create('invalid .SNA');
    st.ReadBuffer(state.regs.ir.i, 1);
    st.ReadBuffer(state.regs.hl[1], 2);
    st.ReadBuffer(state.regs.de[1], 2);
    st.ReadBuffer(state.regs.bc[1], 2);
    st.ReadBuffer(state.regs.af[1], 2);
    st.ReadBuffer(state.regs.hl[0], 2);
    st.ReadBuffer(state.regs.de[0], 2);
    st.ReadBuffer(state.regs.bc[0], 2);
    st.ReadBuffer(state.regs.iy, 2); //iy
    st.ReadBuffer(state.regs.ix, 2); //ix
    st.ReadBuffer(b, 1); //imms
    //WriteLn(b);
    state.iff1 := (b and $04) <> 0;
    state.iff2 := state.iff1;
    st.ReadBuffer(state.regs.ir.r, 1);
    st.ReadBuffer(state.regs.af[0], 2);
    st.ReadBuffer(state.regs.sp, 2);
    st.ReadBuffer(state.im, 1);
    state.im := state.im and $03;
    if state.im = 3 then state.im := 0;  //!!!
    st.ReadBuffer(state.border, 1);
    state.border := state.border and $07;

    st.ReadBuffer(tmp[0], 16384); DZSNewPageWithData(state, 5, tmp[0], 16384);
    st.ReadBuffer(tmp[0], 16384); DZSNewPageWithData(state, 2, tmp[0], 16384);
    if st.Size > 49179 then
    begin
      state.model := dzs128k;
      st.ReadBuffer(tmp[0], 16384); // temp

      st.ReadBuffer(state.regs.pc, 2);
      st.ReadBuffer(state.last7FFD, 1);
      st.ReadBuffer(b, 1); state.trdosIn := (b <> 0);
      DZSNewPageWithData(state, state.last7FFD and $07, tmp[0], 16384);
      for f := 0 to 7 do
      begin
        if (f = 2) or (f = 5) or (f = (state.last7FFD and $07)) then continue;
        st.ReadBuffer(tmp[0], 16384); DZSNewPageWithData(state, f, tmp[0], 16384);
      end;
    end
    else
    begin
      state.model := dzs48k;
      st.ReadBuffer(tmp[0], 16384); DZSNewPageWithData(state, 0, tmp[0], 16384);

      // restore PC from the stack (originally performs reti)
      if (state.regs.sp.w < 16384) or (state.regs.sp.w = 65535) then raise Exception.Create('invalid .SNA');
      state.regs.pc.l :=
        DZSPeek(state, bnk[(state.regs.sp.w shr 14) and $03], (state.regs.sp.w and $3FFF));
      Inc(state.regs.sp.w);
      state.regs.pc.h :=
        DZSPeek(state, bnk[(state.regs.sp.w shr 14) and $03], (state.regs.sp.w and $3FFF));
      Inc(state.regs.sp.w);
    end;
  except
    DZSClearStateRec(state);
    raise;
  end;
end;


// FIXME! deal with >128k machines!
procedure DZSLoadZ80FromStream (st: TStream; var state: PDZSZ80State);
const
  bnk: array [0..3] of Byte = (0, 5, 2, 0);
var
  b, flgz: Byte;
  buf, mem: packed array of Byte;
  f, sstate: Integer;
  hdrExtraLen: Integer;

  procedure SetModel (mdl: Byte; v2: Boolean);
  begin
    if v2 then
    begin
      if mdl = 3 then mdl := 4
      else if mdl = 4 then mdl := 5
      else if mdl > 4 then raise Exception.Create('Z80v2: invalid machine type');
    end;

    case mdl of
      0, 1, 2, 3: state.model := dzs48k;
      // 0: 48k; 1: 48k with interface1; 2: 48k with samram; 3: 48k with MGT(?)
      4, 5, 6: state.model := dzs128k;
      // 4: 128k; 5: 128k with interface1; 6: 128k with MGT(?)
      7: state.model := dzsPlus3;
      //8: +3 xzx error
      9: state.model := dzsPentagon;
      10: state.model := dzsScorpion;
      //11: ???
      12: state.model := dzsPlus2;
      13: state.model := dzsPlus2a;
      //14: tc2048
      //15: tc2068
      //128: ts2048
      else raise Exception.Create('Z80v2: unknown machine type');
    end;
  end;

  procedure LoadV2Hdr ();
  var
    btemp, x: Byte;
    mdl: Byte;
  begin
    st.ReadBuffer(state.regs.pc.w, 2);
    st.ReadBuffer(mdl, 1); SetModel(mdl, true);
    st.ReadBuffer(x, 1); if state.model > dzs48k then state.last7FFD := x;
    st.ReadBuffer(btemp, 1); // ??? timex & so on?
    st.ReadBuffer(btemp, 1); // another timex, spectaculator extensions. fuck 'em off.
    // read AY/AM state
    state.ayType := dzsAY;
    st.ReadBuffer(state.lastFFFD, 1);
    st.ReadBuffer(state.psg[0], 16);
    // 23 bytes read
  end;

  procedure LoadV3Hdr ();
  var
    btemp, x: Byte;
    mdl: Byte;
    f, qst: Integer;
  begin
    st.ReadBuffer(state.regs.pc.w, 2);
    st.ReadBuffer(mdl, 1); SetModel(mdl, false);
    st.ReadBuffer(x, 1); if state.model > dzs48k then state.last7FFD := x;
    st.ReadBuffer(btemp, 1); // ??? timex & so on?
    st.ReadBuffer(btemp, 1); // another timex, spectaculator extensions. fuck 'em off.
    // read AY/AM state
    state.ayType := dzsAY;
    st.ReadBuffer(state.lastFFFD, 1);
    st.ReadBuffer(state.psg[0], 16);
    // 23 bytes read

    // this is correct, even if it does disagree with Z80 v3.05's `z80dump'. taken from libspectrum
    st.ReadBuffer(btemp, 1); qst := btemp;
    st.ReadBuffer(btemp, 1); qst := qst+btemp*256;
    //???st.ReadBuffer(btemp, 1); qst := qst+(((btemp+1) mod 4)+1 * (1773400 div 4)); //???!
    st.ReadBuffer(btemp, 1); qst := qst+btemp*65536;
    if (state.model <> dzsUnknown) and (qst < 1773400) then state.tStates := qst;
    // 26 bytes read. skip others
    for f := 27 to 54 do st.ReadBuffer(btemp, 1);
    if hdrExtraLen = 55 then
    begin
      // +3 memory port
      st.ReadBuffer(btemp, 1); state.last1FFD := btemp;
    end;
  end;

  procedure LoadBlocksV2V3 ();
  var
    f, c, blklen: Integer;
    b, pageNo: Byte;
    buf: packed array [0..16383] of Byte;
    compBuf, ubuf: packed array of Byte;
    slt: array of record id, lid: Word; size: Integer; end;
    sign: packed array [0..2] of Char;
  begin
    SetLength(compBuf, 0); slt := nil; ubuf := nil;
    while st.Position < st.Size do
    begin
      st.ReadBuffer(b, 1); blklen := b;
      st.ReadBuffer(b, 1); blklen := blklen+b*256;
      st.ReadBuffer(pageNo, 1);
      if pageNo >= 18 then raise Exception.Create('Z80v2/3: invalid page number');

      if blklen = 0 then
      begin
        // it can be .SLT, try to load it
        st.ReadBuffer(sign[0], 3);
        if sign <> 'SLT' then exit; // not an .SLT
        repeat
          SetLength(slt, Length(slt)+1);
          st.ReadBuffer(slt[High(slt)].id, 2);
          st.ReadBuffer(slt[High(slt)].lid, 2);
          st.ReadBuffer(slt[High(slt)].size, 4);
        until (slt[High(slt)].id = 0) and (slt[High(slt)].lid = 0) and (slt[High(slt)].size = 0);
        //*WriteLn('SLT blocks: ', High(slt));
        SetLength(ubuf, 65536);
        for f := 0 to High(slt) do
        begin
          if slt[f].size < 0 then break;
          if slt[f].id <> 1 then
          begin
            // only levels
            st.Position := st.Position+slt[f].size;
            continue;
          end;
          //!MessageBox(0, 'level', '!', 0);
          if Length(compBuf) < slt[f].size then SetLength(compBuf, slt[f].size);
          if slt[f].size > 0 then st.ReadBuffer(compBuf[0], slt[f].size);
          c := Z80RLEUnpackBuffer(compBuf[0], slt[f].size, ubuf[0], Length(ubuf));
          if c > 0 then
          begin
            DZSNewAddonBlock(state, slt[f].lid, ubuf[0], c);
          end;
        end;
        exit;
      end;

      if blklen = $FFFF then
      begin
        // 16384 bytes of uncompressed data
        st.ReadBuffer(buf[0], 16384);
      end
      else
      begin
        // uncompress block
        FillChar(buf[0], SizeOf(buf), 0);
        if Length(compBuf) < blklen then SetLength(compBuf, blklen);
        st.ReadBuffer(compBuf[0], blklen);
        if Z80RLEUnpackBuffer(compBuf[0], blklen, buf[0], SizeOf(buf)) = -1 then
          raise Exception.Create('Z80v2/3: invalid compressed block');
      end;
      if pageNo < 3 then continue; // rom page. throw away
      if pageNo = 11 then continue; // miltiface rom (unless this is scorpion or so). throw away
      if pageNo > 11 then raise Exception.Create('Z80v2/3: unsupported memory page');
      if state.model = dzs48k then
      begin
        // renumber blocks slightly
        if pageNo = 3 then continue; // rom page. throw away
        if PageNo = 4 then pageNo := 2
        else if PageNo = 5 then pageNo := 0
        else if pageNo = 8 then pageNo := 5
        else continue// throw away. Exception.Create('Z80v2/3: invalid 48k page!');
      end
      else // 128k
      begin
        Dec(pageNo, 3);
      end;
      DZSNewPageWithData(state, pageNo, buf[0], 16384);
    end;
  end;

begin
  DZSClearStateRec(state);
  try
    st.ReadBuffer(state.regs.af[0].a, 1);
    st.ReadBuffer(state.regs.af[0].f, 1);
    st.ReadBuffer(state.regs.bc[0].w, 2);
    st.ReadBuffer(state.regs.hl[0].w, 2);
    st.ReadBuffer(state.regs.pc.w, 2);
    st.ReadBuffer(state.regs.sp.w, 2);
    st.ReadBuffer(state.regs.ir.i, 1);
    st.ReadBuffer(state.regs.ir.r, 1);
    state.regs.ir.r := state.regs.ir.r and $7F; //???
    st.ReadBuffer(flgz, 1);
    if flgz = 255 then flgz := 1;
    state.regs.ir.r := state.regs.ir.r or ((flgz and $01) shl 7);
    state.border := (flgz and $0E) shr 1;
    st.ReadBuffer(state.regs.de[0].w, 2);
    st.ReadBuffer(state.regs.bc[1].w, 2);
    st.ReadBuffer(state.regs.de[1].w, 2);
    st.ReadBuffer(state.regs.hl[1].w, 2);
    st.ReadBuffer(state.regs.af[1].a, 1);
    st.ReadBuffer(state.regs.af[1].f, 1);
    st.ReadBuffer(state.regs.iy.w, 2); // iy
    st.ReadBuffer(state.regs.ix.w, 2); // ix
    st.ReadBuffer(b, 1); state.iff1 := (b <> 0);
    st.ReadBuffer(b, 1); state.iff2 := (b <> 0);
    st.ReadBuffer(b, 1);
    state.im := (b and $03); if state.im = 3 then raise Exception.Create('Z80: invalid IM mode');

    if state.regs.pc.w = 0 then // v2, v3
    begin
      // read extra length
      st.ReadBuffer(b, 1); hdrExtraLen := b;
      st.ReadBuffer(b, 1); hdrExtraLen := hdrExtraLen+b*256;

      if hdrExtraLen = 23 then LoadV2Hdr()
      else if (hdrExtraLen = 54) or (hdrExtraLen = 55) then LoadV3Hdr()
      else raise Exception.Create('Z80: invalid file (can''t determine version)');

      LoadBlocksV2V3();
    end
    else
    begin
      state.model := dzs48k; // always 48k, header is exactly 30 bytes, that was already read %-)
      if (flgz and $20) = 0 then
      begin
        // uncompressed, just read memory bankz
        SetLength(mem, 16384);
        st.ReadBuffer(mem[0], 16384); DZSNewPageWithData(state, 5, mem[0], 16384);
        st.ReadBuffer(mem[0], 16384); DZSNewPageWithData(state, 2, mem[0], 16384);
        st.ReadBuffer(mem[0], 16384); DZSNewPageWithData(state, 0, mem[0], 16384);
      end
      else
      begin
        // load & unpack
        if st.Size-st.Position > 192*1024 then SetLength(buf, 192*1024) else SetLength(buf, st.Size-st.Position);
        // 192 kb is the worst case (no 48k snapshot can be bigger than 192 kb!)
        // there must be at least four bytes (end marker) for valid Z80v1
        if Length(buf) < 4 then raise Exception.Create('Z80v1: invalid (too small)');
        st.ReadBuffer(buf[0], Length(buf));

        // search end marker
        // code inspired by libspectrum %-)
        sstate := 0; f := 0;
        while (sstate <> 4) and (f < Length(buf)) do
        begin
          case sstate of
            0: if buf[f] = 0 then sstate := 1 else sstate := 0;
            1: if buf[f] = 0 then sstate := 1 else if buf[f] = $ED then sstate := 2 else sstate := 0;
            2: if buf[f] = 0 then sstate := 1 else if buf[f] = $ED then sstate := 3 else sstate := 0;
            3: if buf[f] = 0 then sstate := 4 else sstate := 0;
          end;
          Inc(f);
        end;
        // check if everything is OK
        if sstate <> 4 then raise Exception.Create('Z80v1: invalid (no eof marker)');
        // now f holds the length of the compressed data+4
        // so, decompress it!
        if f > 4 then
        begin
          SetLength(mem, $C000); FillChar(mem[0], Length(mem), 0);
          f := Z80RLEUnpackBuffer(buf[0], f-4, mem[0], Length(mem));
          if f = -1 then raise Exception.Create('Z80v1: invalid (too big)');
          // copy uncompressed data
          DZSNewPageWithData(state, 5, mem[0], 16384);
          DZSNewPageWithData(state, 2, mem[16384], 16384);
          DZSNewPageWithData(state, 0, mem[32768], 16384);
        end;
      end;
    end;
  except
    DZSClearStateRec(state);
    raise;
  end;
end;

procedure DZSSaveSNA (const fName: string; var state: PDZSZ80State);
var
  st: TStream;
begin
  st := CreateFileStream(fName);//TFileStream.Create(fName, fmCreate);
  try
    try
      DZSSaveSNAToStream(st, state);
    finally
      st.Free();
    end;
  except
    SysUtils.DeleteFile(fName);
    raise;
  end;
end;

procedure DZSSaveZ80 (const fName: string; const state: PDZSZ80State; compression: Byte = 1);
var
  st: TStream;
begin
  st := CreateFileStream(fName);//TFileStream.Create(fName, fmCreate);
  try
    try
      DZSSaveZ80ToStream(st, state, compression);
    finally
      st.Free();
    end;
  except
    SysUtils.DeleteFile(fName);
    raise;
  end;
end;

procedure DZSSaveSNAToStream (st: TStream; var state: PDZSZ80State);
const
  bnk: array [0..3] of Byte = (0, 5, 2, 0);
var
  f: Integer;
  b: Byte;
  osp: Word;
  osm0, osm1: Byte;

  procedure WritePage (pg: byte);
  var
    mem: packed array [0..16383] of Byte;
    blk: PDZSMemBlock;
  begin
    FillChar(mem[0], 16384, 0);
    blk := DZSFindPage(state, pg);
    if blk <> nil then
    begin
      if Length(blk.data) > 16384 then Move(blk.data[0], mem[0], Length(blk.data))
      else Move(blk.data[0], mem[0], 16384);
    end;
    st.WriteBuffer(mem[0], 16384);
  end;

begin
  if state.model < dzs48k then raise Exception.Create('SAVESNA: invalid speccy model');

  if state.model = dzs48k then
  begin
    // save PC & memory
    osp := state.regs.sp.w;
    osm0 := DZSPeek(state, bnk[((osp-1) shr 14) and $03], ((osp-1) and $3FFF));
    osm1 := DZSPeek(state, bnk[((osp-2) shr 14) and $03], ((osp-2) and $3FFF));
  end
  else
  begin
    osp := 0; osm0 := 0; osm1 := 0;
  end;

  try
    if state.model = dzs48k then
    begin
      // store PC on the stack
      Dec(state.regs.sp.w);
      DZSPoke(state, bnk[(state.regs.sp.w shr 14) and $03], (state.regs.sp.w and $3FFF), state.regs.pc.h);
      Dec(state.regs.sp.w);
      DZSPoke(state, bnk[(state.regs.sp.w shr 14) and $03], (state.regs.sp.w and $3FFF), state.regs.pc.l);
    end;

    st.WriteBuffer(state.regs.ir.i, 1);
    st.WriteBuffer(state.regs.hl[1], 2);
    st.WriteBuffer(state.regs.de[1], 2);
    st.WriteBuffer(state.regs.bc[1], 2);
    st.WriteBuffer(state.regs.af[1], 2);
    st.WriteBuffer(state.regs.hl[0], 2);
    st.WriteBuffer(state.regs.de[0], 2);
    st.WriteBuffer(state.regs.bc[0], 2);
    st.WriteBuffer(state.regs.iy, 2); //iy
    st.WriteBuffer(state.regs.ix, 2); //ix

    if state.iff1 then b := $04 else b := 0;
    st.WriteBuffer(b, 1); //imms
    st.WriteBuffer(state.regs.ir.r, 1);
    st.WriteBuffer(state.regs.af[0], 2);
    st.WriteBuffer(state.regs.sp, 2);
    st.WriteBuffer(state.im, 1);
    st.WriteBuffer(state.border, 1);
    WritePage(5);
    WritePage(2);

    if state.model > dzs48k then
    begin
      WritePage(state.last7FFD and $07); // current screen

      st.WriteBuffer(state.regs.pc, 2);
      st.WriteBuffer(state.last7FFD, 1);

      // tr-dos
      if state.trdosIn then b := 1 else b := 0;
      st.WriteBuffer(b, 1);

      for f := 0 to 7 do
      begin
        if (f = 2) or (f = 5) or (f = (state.last7FFD and $07)) then continue;
        WritePage(f);
      end;
    end
    else
    begin
      WritePage(0);
    end;
  finally
    if state.model = dzs48k then
    begin
      // restore PC & memory
      state.regs.sp.w := osp;
      if DZSFindPage(state, bnk[((osp-1) shr 14) and $03]) <> nil then
        DZSPoke(state, bnk[((osp-1) shr 14) and $03], ((osp-1) and $3FFF), osm0);
      if DZSFindPage(state, bnk[((osp-2) shr 14) and $03]) <> nil then
        DZSPoke(state, bnk[((osp-2) shr 14) and $03], ((osp-2) and $3FFF), osm1);
    end;
  end;
end;


// FIXME! deal with >128k machines!
procedure DZSSaveZ80ToStream (st: TStream; const state: PDZSZ80State; compression: Byte = 1);
// v3
const
  bnk: array [0..3] of Byte = (0, 5, 2, 0);
var
  b, flgz: Byte;

  function ModelByte (): Byte;
  begin
    case state.model of
      dzs48k: result := 0;
      dzs128k: result := 4;
      dzsPlus2: result := 12;
      dzsPlus2a: result := 13;
      dzsPlus3, dzsPlus3e: result := 7;
      dzsPentagon: result := 9;
      dzsScorpion: result := 10;
      else raise Exception.Create('DZSSAVEZ80: invalid Speccy model');
    end;
  end;

  procedure SaveV3Hdr ();
  var
    btemp, x: Byte;
    f: Integer;
  begin
    st.WriteBuffer(state.regs.pc.w, 2);
    x := ModelByte(); st.WriteBuffer(x, 1);
    st.WriteBuffer(state.last7FFD, 1);
    btemp := 0;
    st.WriteBuffer(btemp, 1); // ??? timex & so on?
    st.WriteBuffer(btemp, 1); // another timex, spectaculator extensions. fuck 'em off.
    // write AY/AM state
    st.WriteBuffer(state.lastFFFD, 1);
    st.WriteBuffer(state.psg[0], 16);
    // 23 bytes written

    // ticks. unused
    btemp := 0; st.WriteBuffer(btemp, 1);
    st.WriteBuffer(btemp, 1); st.WriteBuffer(btemp, 1);
    // 26 bytes written. skip others
    for f := 27 to 54 do st.WriteBuffer(btemp, 1);
  end;

  procedure SaveBlocksV3 ();
  var
    f, blklen: Integer;
    pageNo: Byte;
    compBuf: packed array of Byte;
    mem: packed array [0..16383] of Byte;
    blk: PDZSMemBlock;
  begin
    SetLength(compBuf, 16383);
    for f := 0 to 7 do
    begin
      if state.model = dzs48k then
      begin
        if (f <> 0) and (f <> 2) and (f <> 5) then continue;
        if f = 2 then pageNo := 4 else
        if f = 0 then pageNo := 5 else pageNo := 8;
      end
      else pageNo := (f+3);

      blk := DZSFindPage(state, f);
      if blk = nil then continue;
      FillChar(mem[0], 16384, 0);
      if Length(blk.data) > 16384 then Move(blk.data[0], mem[0], Length(blk.data))
      else Move(blk.data[0], mem[0], 16384);

      if compression = 0 then blklen := -1
      else blklen := Z80RLEPackBuffer(mem[0], 16384, compBuf[0], 16383);

      st.WriteBuffer(blklen, 2);
      st.WriteBuffer(pageNo, 1);
      if blklen = -1 then st.WriteBuffer(mem[0], 16384)
      else st.WriteBuffer(compBuf[0], blklen);
    end;
  end;

  procedure SaveSLT ();
  var
    sign: packed array [0..5] of Char;
    blk: PDZSMemBlock;
    pk: packed array of Byte;
    i, c: Integer;
    w: Word;
  begin
    blk := state.blocks;
    if blk = nil then exit;
    SetLength(pk, 128*256);
    sign[0] := #0; sign[1] := #0; sign[2] := #0;
    sign[3] := 'S'; sign[4] := 'L'; sign[5] := 'T';
    st.WriteBuffer(sign[0], 6);
    // write directory
    while blk <> nil do
    begin
      c := Z80RLEPackBuffer(blk.data[0], Length(blk.data), pk[0], Length(pk));
      if c = -1 then raise Exception.Create('invalid SLT block data');

      w := 1; st.WriteBuffer(w, 2); // level
      st.WriteBuffer(blk.page, 2); // level id
      i := c;
      st.WriteBuffer(i, 4); // size
      blk := blk.next;
    end;
    // end-of-dir marker
    w := 0; i := 0;
    st.WriteBuffer(w, 2); st.WriteBuffer(w, 2); st.WriteBuffer(i, 4);
    // write data
    blk := state.blocks;
    while blk <> nil do
    begin
      if Length(blk.data) > 0 then
      begin
        c := Z80RLEPackBuffer(blk.data[0], Length(blk.data), pk[0], Length(pk));
        if c = -1 then raise Exception.Create('invalid SLT block data');
        st.WriteBuffer(pk[0], c);
      end;
      blk := blk.next;
    end;
  end;

begin
  st.WriteBuffer(state.regs.af[0].a, 1);
  st.WriteBuffer(state.regs.af[0].f, 1);
  st.WriteBuffer(state.regs.bc[0].w, 2);
  st.WriteBuffer(state.regs.hl[0].w, 2);
  b := 0; st.WriteBuffer(b, 1); st.WriteBuffer(b, 1); // pc=0
  st.WriteBuffer(state.regs.sp.w, 2);
  st.WriteBuffer(state.regs.ir.i, 1);
  b := state.regs.ir.r and $7F;
  st.WriteBuffer(b, 1);
  flgz := (state.border and $07) shl 1; if state.regs.ir.r > $7F then flgz := flgz or 1;
  st.WriteBuffer(flgz, 1);
  st.WriteBuffer(state.regs.de[0].w, 2);
  st.WriteBuffer(state.regs.bc[1].w, 2);
  st.WriteBuffer(state.regs.de[1].w, 2);
  st.WriteBuffer(state.regs.hl[1].w, 2);
  st.WriteBuffer(state.regs.af[1].a, 1);
  st.WriteBuffer(state.regs.af[1].f, 1);
  st.WriteBuffer(state.regs.iy.w, 2); // iy
  st.WriteBuffer(state.regs.ix.w, 2); // ix
  if state.iff1 then b := 1 else b := 0;
  st.WriteBuffer(b, 1);
  if state.iff2 then b := 1 else b := 0;
  st.WriteBuffer(b, 1);
  st.WriteBuffer(state.im, 1);

  // write extra length
  b := 54; st.WriteBuffer(b, 1);
  b := 0; st.WriteBuffer(b, 1);
  SaveV3Hdr();
  SaveBlocksV3();
  SaveSLT();
end;


end.
