// Wildcard matching
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// this code is meant to allow wildcard pattern matches. tt is VERY useful
// for matching filename wildcard patterns. tt allows unix grep-like pattern
// comparisons, for instance:
//
//       ?       Matches any single characer
//       +       Matches any single characer or nothing
//       *       Matches any contiguous characters
//       [abc]   Matches a or b or c at that position
//       [!abc]  Matches anything but a or b or c at that position
//       [a-e]   Matches a through e at that position
//
//       'ma?ch.*'       -Would match match.exe, mavch.dat, march.on, etc
//       'this [e-n]s a [!zy]est' -Would match 'this is a test', but would
//                                 not match 'this as a yest'
//
{$INCLUDE options.inc}
{.$DEFINE USE_WIDE_STRINGS}
unit wildMatchU;

interface

uses
  Windows; // CharLowerBuffW


const
  // character defines
  WILD_CHAR_ESCAPE         = '\';
  WILD_CHAR_SINGLE         = '?';
  WILD_CHAR_SINGLE_OR_NONE = '+';
  WILD_CHAR_MULTI          = '*';
  WILD_CHAR_RANGE_OPEN     = '[';
  WILD_CHAR_RANGE          = '-';
  WILD_CHAR_RANGE_CLOSE    = ']';
  WILD_CHAR_RANGE_NOT      = '!';


type
{$IFDEF USE_WIDE_STRINGS}
  TWildString = WideString;
  TWildChar = WideChar;
{$ELSE}
  TWildString = AnsiString;
  TWildChar = AnsiChar;
{$ENDIF}


function WildMatch (pattern, text: TWildString; caseSens: Boolean): Boolean;
function WildListMatch (wildList, text: TWildString; caseSens: Boolean;
  delChar: TWildChar = ':'): Integer;
function HasWildcards (const pattern: TWildString): Boolean;

implementation


function HasWildcards (const pattern: TWildString): Boolean;
begin
  result := (Pos(WILD_CHAR_ESCAPE, pattern) <> 0) or
    (Pos(WILD_CHAR_SINGLE, pattern) <> 0) or
    (Pos(WILD_CHAR_SINGLE_OR_NONE, pattern) <> 0) or
    (Pos(WILD_CHAR_MULTI, pattern) <> 0) or
    (Pos(WILD_CHAR_RANGE_OPEN, pattern) <> 0);
end;

function MatchMask (const pattern: TWildString; p, pend: Integer;
  const text: TWildString; t, tend: Integer): Boolean;
var
  rangeStart, rangeEnd: TWildChar;
  rangeNot, rangeMatched: Boolean;
  ch: TWildChar;
begin
  // sanity checks
  if (pend < 0) or (pend > Length(pattern)) then pend := Length(pattern);
  if (tend < 0) or (tend > Length(text)) then tend := Length(text);
  if t < 1 then t := 1;
  if p < 1 then p := 1;
  while p <= pend do
  begin
    if t > tend then
    begin
      // no more text. check if there's no more chars in pattern (except "*" & "+")
      while (p <= pend) and
            ((pattern[p] = WILD_CHAR_MULTI) or
             (pattern[p] = WILD_CHAR_SINGLE_OR_NONE)) do Inc(p);
      result := (p > pend);
      exit;
    end;
    case pattern[p] of
      WILD_CHAR_SINGLE: ;
      WILD_CHAR_ESCAPE:
        begin
          Inc(p);
          if p > pend then result := false else result := (pattern[p] = text[t]);
          if not result then exit;
        end;
      WILD_CHAR_RANGE_OPEN:
        begin
          result := false;
          Inc(p); if p > pend then exit; // sanity check
          rangeNot := (pattern[p] = WILD_CHAR_RANGE_NOT);
          if rangeNot then begin Inc(p); if p > pend then exit; {sanity check} end;
          if pattern[p] = WILD_CHAR_RANGE_CLOSE then exit; // sanity check
          ch := text[t]; // speed reasons
          rangeMatched := false;
          repeat
            if p > pend then exit; // sanity check
            rangeStart := pattern[p];
            if rangeStart = WILD_CHAR_RANGE_CLOSE then break;
            Inc(p); if p > pend then exit; // sanity check
            if pattern[p] = WILD_CHAR_RANGE then
            begin
              Inc(p); if p > pend then exit; // sanity check
              rangeEnd := pattern[p]; Inc(p);
              if rangeStart < rangeEnd then
              begin
                rangeMatched := (ch >= rangeStart) and (ch <= rangeEnd);
              end
              else rangeMatched := (ch >= rangeEnd) and (ch <= rangeStart);
            end
            else rangeMatched := (ch = rangeStart);
          until rangeMatched;
          if rangeNot = rangeMatched then exit;

          // skip the rest or the range
          while (p <= pend) and (pattern[p] <> WILD_CHAR_RANGE_CLOSE) do Inc(p);
          if p > pend then exit; // sanity check
        end;
      WILD_CHAR_SINGLE_OR_NONE:
        begin
          Inc(p);
          result := MatchMask(pattern, p, pend, text, t, tend);
          if not result then result := MatchMask(pattern, p, pend, text, t+1, tend);
          exit;
        end;
      WILD_CHAR_MULTI:
        begin
          while (p <= pend) and (pattern[p] = WILD_CHAR_MULTI) do Inc(p);
          result := (p > pend); if result then exit;
          while not result and (t <= tend) do
          begin
            result := MatchMask(pattern, p, pend, text, t, tend);
            Inc(t);
          end;
          exit;
        end;
      else result := (pattern[p] = text[t]); if not result then exit;
    end;
    Inc(p); Inc(t);
  end;
  result := (t > tend);
end;


function WildMatch (pattern, text: TWildString; caseSens: Boolean): Boolean;
begin
  if not caseSens then
  begin
    if pattern <> '' then
    begin
      UniqueString(pattern);
      {$IFDEF USE_WIDE_STRINGS}
      CharLowerBuffW(@(pattern[1]), Length(pattern));
      {$ELSE}
      CharLowerBuffA(@(pattern[1]), Length(pattern));
      {$ENDIF}
    end;
    if text <> '' then
    begin
      UniqueString(text);
      {$IFDEF USE_WIDE_STRINGS}
      CharLowerBuffW(@(text[1]), Length(text));
      {$ELSE}
      CharLowerBuffA(@(text[1]), Length(text));
      {$ENDIF}
    end;
  end;
  result := MatchMask(pattern, 1, -1, text, 1, -1);
end;

function WildListMatch (wildList, text: TWildString; caseSens: Boolean;
  delChar: TWildChar = ':'): Integer;
var
  s, e: Integer;
begin
  if not caseSens then
  begin
    if wildList <> '' then
    begin
      UniqueString(wildList);
      {$IFDEF USE_WIDE_STRINGS}
      CharLowerBuffW(@(wildList[1]), Length(wildList));
      {$ELSE}
      CharLowerBuffA(@(wildList[1]), Length(wildList));
      {$ENDIF}
    end;
    if text <> '' then
    begin
      UniqueString(text);
      {$IFDEF USE_WIDE_STRINGS}
      CharLowerBuffW(@(text[1]), Length(text));
      {$ELSE}
      CharLowerBuffA(@(text[1]), Length(text));
      {$ENDIF}
    end;
  end;
  result := 0; s := 1;
  while s <= Length(wildList) do
  begin
    e := s; while e <= Length(wildList) do
    begin
      if wildList[e] = WILD_CHAR_RANGE_OPEN then
      begin
        while (e <= Length(wildList)) and (wildList[e] <> WILD_CHAR_RANGE_CLOSE) do
          Inc(e);
      end;
      if wildList[e] = delChar then break;
      Inc(e);
    end;
    if s < e then
    begin
      if MatchMask(wildList, s, e-1, text, 1, -1) then exit;
    end;
    Inc(result);
    s := e+1;
  end;
  result := -1;
end;


end.
