// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// handling TZX tapes
// Lunter's .tap will be converted to TZX anyway by the loader
// some aspects of converting .TAP to "pulsed" was taken from UnrealSpeccy
//
{.$DEFINE VERBOSE}
{$M-}
unit dspTZXU;

interface

uses
  SysUtils, Classes;


type
  ETZXError = class (Exception);

  TZXPulseInfoArray = array of Integer;
  TZXByteArray = packed array of Byte;
  TZXWordArray = packed array of Word;

  TZXTapeBlock = class
  public
    num: Integer;
    next, prev: TZXTapeBlock;
  end;

  TZXBlockStandard = class (TZXTapeBlock)
  public
    pause: Word; // pause after this block in ms
    data: TZXByteArray;
  end;

  TZXBlockTurbo = class (TZXTapeBlock)
  public
    pilotT: Word;
    sync1T: Word;
    sync2T: Word;
    zeroBitT: Word;
    oneBitT: Word;
    pilotLen: Word;
    bitsInLastByte: Byte; // from 7 till 0, unused=0
    pause: Word; // pause after this block in ms
    data: TZXByteArray;
  end;

  TZXBlockPureTone = class (TZXTapeBlock)
  public
    length: Word;
    toneT: Word;
  end;

  TZXBlockPulseSeq = class (TZXTapeBlock)
  public
    data: TZXWordArray;
  end;

  TZXBlockPureData = class (TZXTapeBlock)
  public
    zeroBitT: Word;
    oneBitT: Word;
    bitsInLastByte: Byte; // from 7 till 0, unused=0
    pause: Word; // pause after this block in ms
    data: TZXByteArray;
  end;

  TZXBlockDirect = class (TZXTapeBlock)
  public
    tPerSample: Word; // t-states per one bit of data
    bitsInLastByte: Byte; // from 7 till 0, unused=0
    pause: Word; // pause after this block in ms
    data: TZXByteArray; // each sample=1 bit. plays from 7 till 0
  end;

  TZXBlockPause = class (TZXTapeBlock)
  public
    length: Integer; // 0: stop the tape
  end;

  TZXBlockGroupStart = class (TZXTapeBlock)
  public
    name: string;
  end;

  TZXBlockGroupEnd = class (TZXTapeBlock)
  public
    start: TZXBlockGroupStart; // group start block
  end;

  TZXBlockJump = class (TZXTapeBlock)
  public
    dest: TZXTapeBlock;
  end;

  TZXBlockLoopStart = class (TZXTapeBlock)
  public
    count: Integer;
  end;

  TZXBlockLoopEnd = class (TZXTapeBlock)
  public
    start: TZXBlockLoopStart; // loop start block
  end;

  TZXBlockCallSeq = class (TZXTapeBlock)
  public
    seq: array of TZXTapeBlock;
  end;

  TZXBlockCallReturn = class (TZXTapeBlock) // return from "call"
  public
  end;

  TZXBlockSelect = class (TZXTapeBlock)
  public
    selections: array of
      record
        name: string;
        start: TZXTapeBlock;
      end;
  end;

  TZXBlockStopIf48k = class (TZXTapeBlock)
  public
  end;

  TZXBlockDescription = class (TZXTapeBlock)
  public
    text: string;
  end;

  TZXBlockMessage = class (TZXTapeBlock)
  public
    text: string;
    pause: Integer; // is second. 0: wait for keypress
  end;

  TZXBlockArchiveInfo = class (TZXTapeBlock)
  public
    text: array of string;
      // 0: fill title
      // 1: software house / publisher
      // 2: author(s)
      // 3: year of publication
      // 4: language
      // 5: comments
      // there is no less than 6 items in the array
  end;

  TZXBlockHWInfo = class (TZXTapeBlock)
  public
    run: TZXWordArray;
    use: TZXWordArray;
    runNoSpecial: TZXWordArray;
    norun: TZXWordArray;
  end;

  TZXBlockEmuInfo = class (TZXTapeBlock)
  public
    data: packed array [0..7] of Byte;
  end;

  TZXBlockCustomInfo = class (TZXTapeBlock)
  public
    infoName: string;
  end;

  TZXBlockCustomInfoOther = class (TZXBlockCustomInfo)
  public
    data: TZXByteArray;
  end;

  TZXBlockCustomInfoPokes = class (TZXBlockCustomInfo)
  public
    text: string;
    data: array of
      record
        text: string;
        pokes: array of
          record
            page: Integer; // <0: ingore paging
            addr: Word;
            value: Integer; // <0 or >255: user inserts
            original: Integer; // <0 or >255: unknown original
          end;
      end;
  end;

  TZXBlockCustomInfoInstructions = class (TZXBlockCustomInfo)
  public
    text: string;
  end;

  TZXBlockCustomInfoScreen = class (TZXBlockCustomInfo)
  public
    text: string; // '': loading screen
    border: Byte;
    data: packed array [0..6911] of Byte; // Speccy format
  end;

  TZXBlockSnap = class (TZXTapeBlock)
  public
    sType: Integer; // only 0 (z80) & 1 (sna)
    data: TZXByteArray;
  end;


  TZXTape = class
  protected
    // linked block-list
    fFirstBlock: TZXTapeBlock;
    fLastBlock: TZXTapeBlock;
    fCurrentBlock: TZXTapeBlock;

    fCount: Integer;
    fCurBlockNum: Integer;

    fVersionLo: Byte;
    fVersionHi: Byte;

  public
    constructor Create ();
    destructor Destroy (); override;

    procedure Clear ();

    procedure LoadFromStream (st: TStream);

    procedure GoFirst ();
    procedure GoLast ();
    procedure GoPrev ();
    procedure GoNext ();

    function GetBlockByNumber (index: Integer): TZXTapeBlock;
    procedure DeleteBlock (blk: TZXTapeBlock);
    procedure InsertBlock (before, blk: TZXTapeBlock); // nil: make last
    procedure AppendBlock (blk: TZXTapeBlock);

    property FirstBlock: TZXTapeBlock read fFirstBlock;
    property LastBlock: TZXTapeBlock read fLastBlock;
    property CurrentBlock: TZXTapeBlock read fCurrentBlock;
    property Count: Integer read fCount;
    property CurBlockNum: Integer read fCurBlockNum;
    property VersionLo: Byte read fVersionLo;
    property VersionHi: Byte read fVersionHi;
  end;

  TZXRawPulses = class
  protected
    fPulses: TList;
    fTotalTicks: Integer;
    fCount: Integer;
    fCurPos: Integer;
    fMessages: TStringList;

    function GetCurrent (): Integer;
    function GetPulses (index: Integer): Integer;
    procedure SetPosition (value: Integer);

  public
    constructor Create ();
    destructor Destroy (); override;

    procedure Build (src: TZXTape);
    procedure Clear ();

    procedure Rewind ();

    procedure BuildFromPause (blk: TZXBlockPause);
    procedure BuildFromDirect (blk: TZXBlockDirect);
    procedure BuildFromPulses (blk: TZXBlockPulseSeq);
    procedure BuildFromTone (blk: TZXBlockPureTone);
    procedure BuildFromPure (blk: TZXBlockPureData);
    procedure BuildFromTurbo (blk: TZXBlockTurbo);
    procedure BuildFromStd (blk: TZXBlockStandard);
    procedure BuildBlock (const data: TZXByteArray;
      pilotLen, lastB: Integer;
      pilotT, sync1T, sync2T, zeroBitT, oneBitT, pausems: Integer);

    procedure AddPulse (time: Integer);
    procedure AddPulses (time, count: Integer);
    procedure AddPause (ms: Integer);
    procedure AddPause128 ();
    procedure AddPause48 ();
    procedure AddMessage (index: Integer);

    procedure Next (); // advance to the next one
    function ReadPulse (): Integer; // and advance to the next one

    property Pulses [index: Integer]: Integer read GetPulses; default;
    property Current: Integer read GetCurrent;
    property Position: Integer read fCurPos write SetPosition;
    property Count: Integer read fCount;
    property TotalTicks: Integer read fTotalTicks;
    property Messages: TStringList read fMessages;
  end;
  // special pulses:
  //  <=-1000: abs(special)-1000 is the info message number
  //           ACHTUNG! this special does not interrupts the
  //                    current block!
  //  -660: new block data follows
  //  -661: turboblock
  //  -662: pure tone block
  //  -663: pulses block
  //  -664: pure data block
  //  -665: direct block
  //  -666: standard block
  //  -667: pause block
  //  -3: stop the tape if emu is in 128K mode
  //  -2: stop the tape
  //  -1: stop the tape if emu is in 48K mode

  // specials order:
  //  block type (<-660), then -660, then immediately block data
  //  if this is special pause block (-1,-2,-3), then there's no -660
  //  or so. BUT there can be zero-length blocks!


function LoadLunterTapeFromStream (st: TStream): TZXTape;
procedure LunterTapeToTZX (fi, fo: TStream);


implementation


{ TZXTape }
constructor TZXTape.Create ();
begin
  fFirstBlock := nil; fLastBlock := nil; fCurrentBlock := nil;
end;

destructor TZXTape.Destroy ();
begin
  Clear();
end;

procedure TZXTape.Clear ();
var
  b, n: TZXTapeBlock;
begin
  b := fFirstBlock;
  while b <> nil do begin n := b.next; b.Free(); b := n; end;
  fFirstBlock := nil; fLastBlock := nil; fCurrentBlock := nil;
  fVersionLo := 0; fVersionHi := 0;
  fCount := 0; fCurBlockNum := 0;
end;

procedure TZXTape.LoadFromStream (st: TStream);
var
  sign: packed array [0..7] of Char;
  id: packed array [0..15] of Char;
  curN: Integer;
  bt, b: Byte;
  w: Word;
  n, f, c: Integer;
  s: string;
  barr: packed array of Byte;
  blk, bc: TZXTapeBlock;
  b10: TZXBlockStandard;
  b11: TZXBlockTurbo;
  b12: TZXBlockPureTone;
  b13: TZXBlockPulseSeq;
  b14: TZXBlockPureData;
  b15: TZXBlockDirect;
  b20: TZXBlockPause;
  b21: TZXBlockGroupStart;
  b22: TZXBlockGroupEnd;
  b23: TZXBlockJump;
  b24: TZXBlockLoopStart;
  b25: TZXBlockLoopEnd;
  b26: TZXBlockCallSeq;
  b27: TZXBlockCallReturn;
  b28: TZXBlockSelect;
  b2a: TZXBlockStopIf48k;
  b30: TZXBlockDescription;
  b31: TZXBlockMessage;
  b32: TZXBlockArchiveInfo;
  b33: TZXBlockHWInfo;
  b34: TZXBlockEmuInfo;
  b35x: TZXBlockCustomInfoOther;
  b350: TZXBlockCustomInfoPokes;
  b351: TZXBlockCustomInfoInstructions;
  b352: TZXBlockCustomInfoScreen;
  b40: TZXBlockSnap;
  //b5a
begin
  Clear(); barr := nil;
  try
    st.ReadBuffer(sign, SizeOf(sign));
    if sign <> 'ZXTape!'#26 then raise ETZXError.Create('invalid TZX signature');
    st.ReadBuffer(fVersionHi, 1);
    st.ReadBuffer(fVersionLo, 1);
    if fVersionHi <> 1 then raise ETZXError.Create('invalid TZX version');
    if fVersionLo > 13 then WriteLn('TZX WARNING: version 1.', fVersionLo, ' found.'); //!!!

    curN := 0;
    while st.Position < st.Size do
    begin
      st.ReadBuffer(bt, 1);
      {$IFDEF VERBOSE}WriteLn(IntToHex(st.Position-1, 6), ': block type ', IntToHex(bt, 2));{$ENDIF}
      blk := nil;
      try
        case bt of
          $10: // normal block
            begin
              b10 := TZXBlockStandard.Create(); blk := b10;
              st.ReadBuffer(w, 2); b10.pause := w;
              {$IFDEF VERBOSE}WriteLn('tzx normal pause: ', b10.pause);{$ENDIF}

              st.ReadBuffer(w, 2);
              SetLength(b10.data, w);
              if w > 0 then st.ReadBuffer(b10.data[0], w);
            end;
          $11: // turbo block
            begin
              b11 := TZXBlockTurbo.Create(); blk := b11;
              st.ReadBuffer(w, 2); b11.pilotT := w;
              st.ReadBuffer(w, 2); b11.sync1T := w;
              st.ReadBuffer(w, 2); b11.sync2T := w;
              st.ReadBuffer(w, 2); b11.zeroBitT := w;
              st.ReadBuffer(w, 2); b11.oneBitT := w;
              st.ReadBuffer(w, 2); b11.pilotLen := w;
              st.ReadBuffer(b, 1); b11.bitsInLastByte := b;
              st.ReadBuffer(w, 2); b11.pause := w;
              {$IFDEF VERBOSE}WriteLn('tzx turbo pause: ', b11.pause);{$ENDIF}

              n := 0; st.ReadBuffer(n, 3);
              SetLength(b11.data, n);
              if n > 0 then st.ReadBuffer(b11.data[0], n);

              if (b11.pilotLen = 0) and (Length(b11.data) = 0) and (b11.pause = 0) then
                begin blk.Free(); blk := nil; end;
            end;
          $12: // pure tone
            begin
              b12 := TZXBlockPureTone.Create(); blk := b12;
              // ??? reverse order of tone & length?
              st.ReadBuffer(w, 2); b12.toneT := w;
              st.ReadBuffer(w, 2); b12.length := w;
              if b12.length = 0 then begin blk.Free(); blk := nil; end;
            end;
          $13: // pure sequence
            begin
              b13 := TZXBlockPulseSeq.Create(); blk := b13;
              st.ReadBuffer(b, 1);
              SetLength(b13.data, b);
              if b > 0 then st.ReadBuffer(b13.data[0], Integer(b)*2)
              else begin blk.Free(); blk := nil; end;
            end;
          $14: // pure data
            begin
              b14 := TZXBlockPureData.Create(); blk := b14;
              st.ReadBuffer(w, 2); b14.zeroBitT := w;
              st.ReadBuffer(w, 2); b14.oneBitT := w;
              st.ReadBuffer(b, 1); b14.bitsInLastByte := b;
              st.ReadBuffer(w, 2); b14.pause := w;

              n := 0; st.ReadBuffer(n, 3);
              SetLength(b14.data, n);
              if n > 0 then st.ReadBuffer(b14.data[0], n);

              if (Length(b14.data) = 0) and (b14.pause = 0) then
                begin blk.Free(); blk := nil; end;
            end;
          $15: // direct
            begin
              b15 := TZXBlockDirect.Create(); blk := b15;
              st.ReadBuffer(w, 2); b15.tPerSample := w;
              st.ReadBuffer(b, 1); b15.bitsInLastByte := b;
              st.ReadBuffer(w, 2); b15.pause := w;

              st.ReadBuffer(b, 1); f := b;
              st.ReadBuffer(b, 1); f := f+b*256;
              st.ReadBuffer(b, 1); f := f+b*256*256;

              SetLength(b15.data, f);
              if f > 0 then st.ReadBuffer(b15.data[0], f);

              if (Length(b15.data) = 0) and (b15.pause = 0) then
              else begin blk.Free(); blk := nil; end;
            end;
          $16: // C64 ROM, skip it
            begin
              WriteLn('TZX: C64 ROM block');
              st.ReadBuffer(f, 4); // length
              if f < 0 then f := 0;
              st.Seek(f, soFromCurrent);
            end;
          $17: // C64 turbo, skip it
            begin
              WriteLn('TZX: C64 turbo block');
              st.ReadBuffer(f, 4); // length
              if f < 0 then f := 0;
              st.Seek(f, soFromCurrent);
            end;
          $20: // pause
            begin
              b20 := TZXBlockPause.Create(); blk := b20;
              st.ReadBuffer(w, 2); b20.length := w;
            end;
          $21: // group start
            begin
              b21 := TZXBlockGroupStart.Create(); blk := b21;
              st.ReadBuffer(b, 1); SetLength(s, b);
              if b > 0 then st.ReadBuffer(s[1], b);
              b21.name := s;
            end;
          $22: // group end
            begin
              b22 := TZXBlockGroupEnd.Create(); blk := b22;
            end;
          $23: // jump
            begin
              b23 := TZXBlockJump.Create(); blk := b23;
              st.ReadBuffer(b, 1); f := b;
              st.ReadBuffer(b, 1); f := f+b*256;
              if f > 32767 then f := -(65536-f);
              if f = 0 then begin blk.Free(); blk := nil; end
              else b23.dest := Pointer(curN+f); // fix later!
            end;
          $24: // loop start
            begin
              b24 := TZXBlockLoopStart.Create(); blk := b24;
              st.ReadBuffer(w, 2); b24.count := b;
            end;
          $25: // loop end
            begin
              b25 := TZXBlockLoopEnd.Create(); blk := b25;
            end;
          $26: // call sequence
            begin
              b26 := TZXBlockCallSeq.Create(); blk := b26;
              st.ReadBuffer(w, 2); n := b;
              SetLength(b26.seq, n);
              for c := 0 to n-1 do
              begin
                st.ReadBuffer(b, 1); f := b;
                st.ReadBuffer(b, 1); f := f+b*256;
                if f > 32767 then f := -(65536-f);
                b26.seq[c] := Pointer(curN+f);
              end;

              // remove self-calls
              f := 0; while f < Length(b26.seq) do
              begin
                if Integer(b26.seq[f]) = curN then
                begin
                  for c := f+1 to High(b26.seq) do b26.seq[c-1] := b26.seq[c];
                  SetLength(b26.seq, Length(b26.seq)-1);
                end
                else Inc(f);
              end;

              if Length(b26.seq) = 0 then begin blk.Free(); blk := nil; end;
            end;
          $27: // return
            begin
              b27 := TZXBlockCallReturn.Create(); blk := b27;
            end;
          $28: // select
            begin
              //WriteLn('no "selection" blocks!');
              //raise ETZXError.Create('no "selection" blocks!');

              WriteLn('"selection" block');
              b28 := TZXBlockSelect.Create(); blk := b28;
              st.ReadBuffer(w, 2);
              st.ReadBuffer(n, 1); SetLength(b28.selections, n);
              for c := 0 to n-1 do
              begin
                st.ReadBuffer(b, 1); f := b;
                st.ReadBuffer(b, 1); f := f+b*256;
                if f > 32767 then f := -(65536-f);
                b28.selections[c].start := Pointer(curN+f);

                st.ReadBuffer(b, 1); SetLength(s, b);
                if b > 0 then st.ReadBuffer(s[1], b);
                b28.selections[c].name := s;
                WriteLn(' ', Integer(b28.selections[c].start), ': ', s);
              end;
              WriteLn('========================================');

              (*// remove self-pointers
              f := 0; while f < Length(b28.selections) do
              begin
                if Integer(b28.selections[f].start) = curN then
                begin
                  for c := f+1 to High(b28.selections) do b28.selections[c-1] := b28.selections[c];
                  SetLength(b28.selections, Length(b28.selections)-1);
                end
                else Inc(f);
              end;

              if Length(b28.selections) = 0 then begin blk.Free(); blk := nil; end;*)

              blk.Free();
              blk := nil;
            end;
          $2A: // stop if 48k
            begin
              b2a := TZXBlockStopIf48k.Create(); blk := b2a;
            end;
          $30: // text description
            begin
              b30 := TZXBlockDescription.Create(); blk := b30;
              st.ReadBuffer(b, 1); SetLength(s, b);
              if b > 0 then st.ReadBuffer(s[1], b);
              b30.text := s;
              if b30.text = '' then begin blk.Free(); blk := nil; end;
            end;
          $31: // message
            begin
              b31 := TZXBlockMessage.Create(); blk := b31;
              st.ReadBuffer(b, 1); b31.pause := b;
              st.ReadBuffer(b, 1); SetLength(s, b);
              if b > 0 then st.ReadBuffer(s[1], b);
              b31.text := s;
              if b31.text = '' then begin blk.Free(); blk := nil; end;
            end;
          $32: // archive info
            begin
              b32 := TZXBlockArchiveInfo.Create(); blk := b32;
              st.ReadBuffer(w, 2); //FIXME: check length
              st.ReadBuffer(b, 1); n := b;
              SetLength(b32.text, 6); for c := 0 to 5 do b32.text[c] := '';
              for c := 1 to n do
              begin
                st.ReadBuffer(b, 1); f := b;
                st.ReadBuffer(b, 1); SetLength(s, b);
                if b > 0 then st.ReadBuffer(s[1], b);
                if s <> '' then
                begin
                  if (f < 5) or (f = 255) then
                  begin
                    if f = 255 then f := 5;
                    if b32.text[f] <> '' then b32.text[f] := b32.text[f]+#13;
                    b32.text[f] := b32.text[f]+s;
                  end
                  else
                  begin
                    s := chr(f)+s;
                    f := 6;
                    while (f < Length(b32.text)) and (b32.text[f][1] <> s[1]) do Inc(f);
                    if f = Length(b32.text) then
                    begin
                      SetLength(b32.text, Length(b32.text)+1);
                      b32.text[High(b32.text)] := s;
                    end
                    else b32.text[f] := b32.text[f]+#13+Copy(s, 2, Length(s));
                  end;
                end;
              end;

              f := 0;
              for c := 0 to High(b32.text) do
                if b32.text[c] <> '' then begin f := 1; break; end;

              if f = 0 then begin blk.Free(); blk := nil; end;
            end;
          $33: // hardware type
            begin
              b33 := TZXBlockHWInfo.Create(); blk := b33;
              st.ReadBuffer(b, 1); SetLength(barr, b*3);
              if b > 0 then
              begin
                b33.run := nil; b33.use := nil; b33.runNoSpecial := nil; b33.norun := nil;
                n := b; for c := 1 to n do
                begin
                  st.ReadBuffer(w, 2); st.ReadBuffer(b, 1);
                  case b of
                    0:
                      begin
                        SetLength(b33.run, Length(b33.run)+1);
                        b33.run[High(b33.run)] := w;
                      end;
                    1:
                      begin
                        SetLength(b33.use, Length(b33.use)+1);
                        b33.use[High(b33.use)] := w;
                      end;
                    2:
                      begin
                        SetLength(b33.runNoSpecial, Length(b33.runNoSpecial)+1);
                        b33.runNoSpecial[High(b33.runNoSpecial)] := w;
                      end;
                    3:
                      begin
                        SetLength(b33.norun, Length(b33.norun)+1);
                        b33.norun[High(b33.norun)] := w;
                      end;
                  end;
                end;
              end;

              if (Length(b33.run) = 0) and (Length(b33.use) = 0) and
                 (Length(b33.runNoSpecial) = 0) and (Length(b33.norun) = 0) then
                begin blk.Free(); blk := nil; end;
            end;
          $34: // emulator info
            begin
              b34 := TZXBlockEmuInfo.Create(); blk := b34;
              st.ReadBuffer(b34.data[0], 8);
            end;
          $35: // custom info
            begin
              //b35 := TZXBlockEmuInfo.Create(); blk := b34;
              st.ReadBuffer(id, 16);
              st.ReadBuffer(n, 4); // length
              {$IFDEF VERBOSE}WriteLn(id, ': ', n);{$ENDIF}
              if id = 'POKEs           ' then
              begin
                WriteLn('============POKEZ============');
                b350 := TZXBlockCustomInfoPokes.Create(); blk := b350;
                st.ReadBuffer(b, 1); SetLength(s, b);
                if b > 0 then st.ReadBuffer(s[1], b);
                b350.text := s;
                WriteLn('"', b350.text, '"');
                st.ReadBuffer(b, 1); n := b;
                SetLength(b350.data, n);
                for c := 0 to High(b350.data) do
                begin
                  st.ReadBuffer(b, 1); SetLength(s, b);
                  if b > 0 then st.ReadBuffer(s[1], b);
                  b350.data[c].text := s;
                  st.ReadBuffer(b, 1); SetLength(b350.data[c].pokes, b);
                  //{$IFDEF VERBOSE}WriteLn(' "', b350.data[c].text, '"');{$ENDIF}
                  WriteLn(' ', c, ':"', b350.data[c].text, '"');
                  for f := 0 to High(b350.data[c].pokes) do
                  begin
                    st.ReadBuffer(b, 1);
                    b350.data[c].pokes[f].page := (b and $07);
                    if (b and $08) <> 0 then b350.data[c].pokes[f].page := -1;
                    if (b and $10) <> 0 then b350.data[c].pokes[f].value := -1
                    else b350.data[c].pokes[f].value := 0;
                    if (b and $20) <> 0 then b350.data[c].pokes[f].original := -1
                    else b350.data[c].pokes[f].original := 0;
                    st.ReadBuffer(w, 2); b350.data[c].pokes[f].addr := w;
                    st.ReadBuffer(b, 1);
                    if b350.data[c].pokes[f].value <> -1 then b350.data[c].pokes[f].value := b;
                    st.ReadBuffer(b, 1);
                    if b350.data[c].pokes[f].original <> -1 then b350.data[c].pokes[f].original := b;
                    (*{$IFDEF VERBOSE}WriteLn(b350.data[c].pokes[f].page, ':', b350.data[c].pokes[f].addr, ' := ',
                      b350.data[c].pokes[f].value, ' [', b350.data[c].pokes[f].original, ']');{$ENDIF}*)
                    Write('  ');
                    if b350.data[c].pokes[f].page <> -1 then Write(b350.data[c].pokes[f].page, ':');
                    WriteLn('#', IntToHex(b350.data[c].pokes[f].addr, 4),
                      ' := ',
                      b350.data[c].pokes[f].value, ' [', b350.data[c].pokes[f].original, ']');
                  end;
                end;
                WriteLn('============POKEZ============');

                // remove empty pokes
                f := 0; while f < Length(b350.data) do
                begin
                  if Length(b350.data[f].pokes) = 0 then
                  begin
                    for c := f+1 to High(b350.data) do b350.data[c-1] := b350.data[c];
                    SetLength(b350.data, Length(b350.data)-1);
                  end
                  else Inc(f);
                end;

                f := 0;
                for c := 0 to High(b350.data) do
                  if Length(b350.data[c].pokes) > 0 then begin f := 1; break; end;

                if f = 0 then begin blk.Free(); blk := nil; end;
              end
              else if id = 'Instructions    ' then
              begin
                b351 := TZXBlockCustomInfoInstructions.Create(); blk := b351;
                SetLength(s, n);
                if n > 0 then st.ReadBuffer(s[1], n)
                else begin blk.Free(); blk := nil; end;
              end
              else if id = 'Spectrum Screen ' then
              begin
                b352 := TZXBlockCustomInfoScreen.Create(); blk := b352;
                st.ReadBuffer(b, 1); SetLength(s, b);
                if b > 0 then st.ReadBuffer(s[1], b);
                b352.text := s;
                st.ReadBuffer(b, 1); b352.border := b;
                st.ReadBuffer(b352.data[0], 6912);
              end
              else if n > 0 then
              begin
                b35x := TZXBlockCustomInfoOther.Create(); blk := b35x;
                SetLength(b35x.data, n);
                st.ReadBuffer(b35x.data[0], n);
              end;
            end;
          $40: // snapshot
            begin
              b40 := TZXBlockSnap.Create(); blk := b40;
              st.ReadBuffer(b, 1); b40.sType := b;
              st.ReadBuffer(w, 2); SetLength(b40.data, w);
              if w > 0 then st.ReadBuffer(b40.data[0], w)
              else begin blk.Free(); blk := nil; end;
            end;
          $5A:
            begin
              st.ReadBuffer(id[0], 9);
            end;
          else
            WriteLn('unknown block: $', IntToHex(bt, 2), ' at $',
              IntToHex(st.Position-1, 8));
            break;
        end;

        // add loaded block
        if blk <> nil then
        begin
          {$IFDEF VERBOSE}
          WriteLn('block no ', curN, ', type: "', blk.ClassName, '"');
          {$ENDIF}
          blk.num := curN; AppendBlock(blk);
          blk := nil; // block added ok
        end;

        Inc(curN);
      finally
        blk.Free();
      end;
    end;
    WriteLn('TZX: ', curN, ' block(s) loaded ok');

    // now process jumps and calls
    blk := fFirstBlock;
    while blk <> nil do
    begin
      if blk is TZXBlockJump then
      begin
        b23 := TZXBlockJump(blk);
        f := Integer(b23.dest);
        bc := fFirstBlock;
        while (bc <> nil) and (bc.num <> f) do bc := bc.next;
        if bc <> nil then b23.dest := bc
        else
        begin
          bc := blk.next; DeleteBlock(blk); blk := bc; continue;
        end;
      end
      else if blk is TZXBlockLoopEnd then
      begin
        b25 := TZXBlockLoopEnd(blk);
        bc := blk.prev;
        while (bc <> nil) and not (bc is TZXBlockLoopStart) do bc := bc.prev;
        if bc <> nil then b25.start := TZXBlockLoopStart(bc)
        else
        begin
          bc := blk.next; DeleteBlock(blk); blk := bc; continue;
        end;
      end
      else if blk is TZXBlockCallSeq then
      begin
        b26 := TZXBlockCallSeq(blk);
        c := 0;
        while c < Length(b26.seq) do
        begin
          f := Integer(b26.seq[c]);
          bc := fFirstBlock;
          while (bc <> nil) and (bc.num <> f) do bc := bc.next;
          if bc <> nil then begin b26.seq[c] := bc; Inc(c); end
          else
          begin
            for f := c+1 to High(b26.seq) do b26.seq[f-1] := b26.seq[f];
            SetLength(b26.seq, Length(b26.seq)-1);
          end;
        end;

        if Length(b26.seq) = 0 then
        begin
          bc := blk.next; DeleteBlock(blk); blk := bc; continue;
        end;
      end
      else if blk is TZXBlockSelect then
      begin
        b28 := TZXBlockSelect(blk);
        c := 0;
        while c < Length(b28.selections) do
        begin
          f := Integer(b28.selections[c].start);
          bc := fFirstBlock;
          while (bc <> nil) and (bc.num <> f) do bc := bc.next;
          if bc <> nil then begin b28.selections[c].start := bc; Inc(c); end
          else
          begin
            for f := c+1 to High(b28.selections) do b28.selections[f-1] := b28.selections[f];
            SetLength(b28.selections, Length(b28.selections)-1);
          end;
        end;

        if Length(b28.selections) = 0 then
        begin
          bc := blk.next; DeleteBlock(blk); blk := bc; continue;
        end;
      end;

      blk := blk.next;
    end;
  except
    Clear();
    raise;
  end;
end;

procedure TZXTape.DeleteBlock (blk: TZXTapeBlock);
begin
  if fCount > 0 then Dec(fCount);
  if fCurrentBlock = blk then fCurrentBlock := fCurrentBlock.next;
  if blk.prev = nil then
  begin
    ASSERT(fFirstBlock = blk);
    fFirstBlock := blk.next;
    if fFirstBlock <> nil then fFirstBlock.prev := nil;
  end
  else blk.prev.next := blk.next;
  if blk.next = nil then
  begin
    ASSERT(fLastBlock = blk);
    fLastBlock := blk.prev;
    if fLastBlock <> nil then fLastBlock.next := nil;
  end
  else blk.next.prev := blk.prev;
  blk.Free();
  if fCurrentBlock = nil then
  begin
    fCurrentBlock := fLastBlock;
    fCurBlockNum := fCount-1;
  end;
end;

procedure TZXTape.InsertBlock (before, blk: TZXTapeBlock);
// nil: make last
begin
  Inc(fCount);
  if before = nil then AppendBlock(blk)
  else
  begin
    blk.next := before; blk.prev := before.prev; before.prev := blk;
    if before.prev = nil then fFirstBlock := blk
    else before.prev.next := blk;
  end;
end;

procedure TZXTape.AppendBlock (blk: TZXTapeBlock);
begin
  Inc(fCount);
  blk.prev := fLastBlock; blk.next := nil;
  if fLastBlock = nil then
  begin
    fFirstBlock := blk; fCurrentBlock := blk;
  end
  else fLastBlock.next := blk;
  fLastBlock := blk;
end;

procedure TZXTape.GoPrev ();
begin
  if (fCurrentBlock <> nil) and (fCurrentBlock.prev <> nil) then
  begin
    fCurrentBlock := fCurrentBlock.prev;
    Dec(fCurBlockNum);
  end;
end;

procedure TZXTape.GoNext ();
begin
  if fCurrentBlock <> nil then
  begin
    fCurrentBlock := fCurrentBlock.next;
    Inc(fCurBlockNum);
  end;
end;

procedure TZXTape.GoFirst ();
begin
  fCurrentBlock := fFirstBlock;
  fCurBlockNum := 0;
end;

procedure TZXTape.GoLast ();
begin
  fCurrentBlock := fLastBlock;
  fCurBlockNum := fCount-1;
end;

function TZXTape.GetBlockByNumber (index: Integer): TZXTapeBlock;
begin
  if index < 0 then begin result := nil; exit; end;
  result := fFirstBlock;
  while (index > 0) and (result <> nil) do
  begin
    result := result.next; Dec(index);
  end;
end;


{ TZXRawPulses }
constructor TZXRawPulses.Create ();
begin
  fPulses := TList.Create();
  fMessages := TStringList.Create();
end;

destructor TZXRawPulses.Destroy ();
begin
  fPulses.Free();
end;

procedure TZXRawPulses.Build (src: TZXTape);
var
  blk: TZXTapeBlock;
begin
  Clear();
  try
    if src <> nil then
    begin
      src.GoFirst();
      while src.CurrentBlock <> nil do
      begin
        blk := src.CurrentBlock;

        if blk is TZXBlockStandard then BuildFromStd(TZXBlockStandard(blk))
        else if blk is TZXBlockTurbo then BuildFromTurbo(TZXBlockTurbo(blk))
        else if blk is TZXBlockPureTone then BuildFromTone(TZXBlockPureTone(blk))
        else if blk is TZXBlockPulseSeq then BuildFromPulses(TZXBlockPulseSeq(blk))
        else if blk is TZXBlockPureData then BuildFromPure(TZXBlockPureData(blk))
        else if blk is TZXBlockDirect then BuildFromDirect(TZXBlockDirect(blk))
        else if blk is TZXBlockPause then BuildFromPause(TZXBlockPause(blk))
        else if blk is TZXBlockStopIf48k then AddPause48()
        else if blk is TZXBlockGroupStart then
        begin
          if TZXBlockGroupStart(blk).name <> '' then
          begin
            AddPulse(-660);
            AddMessage(fMessages.Add(TZXBlockGroupStart(blk).name));
          end;
        end
        else if blk is TZXBlockDescription then
        begin
          if TZXBlockDescription(blk).text <> '' then
          begin
            AddPulse(-660);
            AddMessage(fMessages.Add(TZXBlockDescription(blk).text));
          end;
        end
        else if blk is TZXBlockMessage then
        begin
          if TZXBlockDescription(blk).text <> '' then
          begin
            AddPulse(-660);
            AddMessage(fMessages.Add(TZXBlockMessage(blk).text));
            if TZXBlockMessage(blk).pause = 0 then AddPulse(-2);
          end;
        end;
        src.GoNext();
      end;
    end;
  except
    fPulses.Clear(); fCount := 0; fTotalTicks := 0;
    raise;
  end;
end;

procedure TZXRawPulses.Clear ();
begin
  fPulses.Clear(); fMessages.Clear();
  fCurPos := 0; fCount := 0; fTotalTicks := 0;
end;

procedure TZXRawPulses.Rewind ();
begin
  fCurPos := 0;
end;

function TZXRawPulses.GetCurrent (): Integer;
begin
  if (fCurPos < 0) or (fCurPos >= fPulses.Count) then result := -2
  else result := Integer(fPulses[fCurPos]);
end;

procedure TZXRawPulses.Next ();
begin
  if fCurPos < 0 then fCurPos := 0 else Inc(fCurPos);
  if fCurPos > fPulses.Count then fCurPos := fPulses.Count;
end;

function TZXRawPulses.ReadPulse (): Integer;
begin
  result := GetCurrent(); Next();
end;

function TZXRawPulses.GetPulses (index: Integer): Integer;
begin
  if (index < 0) or (index >= fPulses.Count) then result := -2
  else result := Integer(fPulses[index]);
end;

procedure TZXRawPulses.SetPosition (value: Integer);
begin
  if value < 0 then value := 0
  else if value > fPulses.Count then value := fPulses.Count;
  fCurPos := value;
end;

procedure TZXRawPulses.AddMessage (index: Integer);
begin
  if (index < 0) or (index >= fMessages.Count) then exit;
  AddPulse(-(1000+index));
end;

procedure TZXRawPulses.AddPulse (time: Integer);
begin
  //if time = 0 then exit;
  fPulses.Add(Pointer(time));
  fCount := fPulses.Count;
  if time > 0 then Inc(fTotalTicks, time);
end;

procedure TZXRawPulses.AddPulses (time, count: Integer);
begin
  //if time = 0 then exit;
  while count > 0 do
  begin
    AddPulse(time);
    Dec(count);
  end;
end;

procedure TZXRawPulses.AddPause (ms: Integer);
begin
  if ms <= 0 then AddPulse(-2) else AddPulse(ms*3500);
end;

procedure TZXRawPulses.AddPause48 ();
begin
  AddPulse(-1);
end;

procedure TZXRawPulses.AddPause128 ();
begin
  AddPulse(-3);
end;

procedure TZXRawPulses.BuildBlock (const data: TZXByteArray;
  pilotLen, lastB: Integer;
  pilotT, sync1T, sync2T, zeroBitT, oneBitT, pausems: Integer);
// build "pulsing" with the given timings
// <data> is the data (the ordinary bytes) %-)
var
  f, c, cnt, t: Integer;
begin
  AddPulse(-660); // block data mark

  if lastB > 8 then lastB := 8 else if lastB < 0 then lastB := 0;

  // pilot & sync
  if pilotLen >= 0 then
  begin
    {if pilotT <> 0 then }AddPulses(pilotT, pilotLen);
    // syncs
    {if sync1T <> 0 then }AddPulse(sync1T);
    {if sync2T <> 0 then }AddPulse(sync2T);
  end;

  // data
  for f := 0 to High(data) do
  begin
    if f = High(data) then cnt := lastB-1 else cnt := 7;
    for c := 0 to cnt do
    begin
      if (data[f] and ($80 shr c)) = 0 then t := zeroBitT else t := oneBitT;
      AddPulse(t); AddPulse(t);
    end;
  end;

  if pausems <> 0 then AddPulse(pausems*3500);
end;

procedure TZXRawPulses.BuildFromStd (blk: TZXBlockStandard);
var
  f, pl: Integer;
  s: string;
begin
  AddPulse(-666); // standard block mark
  if (Length(blk.data) = 19) and (blk.data[0] = 0) and (blk.data[1] in [0..3]) then
  begin
    case blk.data[1] of
      0: s := 'PROG: ';
      1: s := 'NUMB: ';
      2: s := 'DATA: ';
      else s := 'CODE: ';
    end;
    for f := 2 to 11 do
    begin
      if chr(blk.data[f]) in [' '..#127] then s := s+chr(blk.data[f]) else s := s+'?';
    end;
    s := Trim(s)+'; len='+IntToStr(PWord(@(blk.data[12]))^);
    if (blk.data[1] = 0) or (blk.data[1] = 3) then
      s := s+'; start='+IntToStr(PWord(@(blk.data[14]))^);
    if blk.data[1] = 0 then
    begin
      s := s+'; prg.len='+IntToStr(PWord(@(blk.data[16]))^);
    end
    else if blk.data[1] <> 3 then
    begin
      s := s+'  var='+chr(blk.data[15]);
    end;
    AddMessage(fMessages.Add(s));
  end;
  if (Length(blk.data) > 0) and (blk.data[0] = 0) then pl := 8064 else pl := 3220;
  BuildBlock(blk.data, pl, 8, 2168, 667, 735, 855, 1710, blk.pause);
end;

procedure TZXRawPulses.BuildFromTurbo (blk: TZXBlockTurbo);
begin
  AddPulse(-661); // turbo block mark
  BuildBlock(blk.data, blk.pilotLen, blk.bitsInLastByte,
    blk.pilotT, blk.sync1T, blk.sync2T, blk.zeroBitT, blk.oneBitT, blk.pause);
end;

procedure TZXRawPulses.BuildFromPure (blk: TZXBlockPureData);
begin
  AddPulse(-664); // pure data block mark
  BuildBlock(blk.data, -1, blk.bitsInLastByte,
    0, 0, 0, blk.zeroBitT, blk.oneBitT, blk.pause);
end;

procedure TZXRawPulses.BuildFromTone (blk: TZXBlockPureTone);
begin
  AddPulse(-662); // pure tone block mark
  AddPulse(-660); // block data mark
  AddPulses(blk.toneT, blk.Length);
end;

procedure TZXRawPulses.BuildFromPulses (blk: TZXBlockPulseSeq);
var
  f: Integer;
begin
  AddPulse(-663); // pulses block mark
  AddPulse(-660); // block data mark
  for f := 0 to High(blk.data) do AddPulse(blk.data[f]);
end;

procedure TZXRawPulses.BuildFromDirect (blk: TZXBlockDirect);
var
  tps, ticks, f, c, cnt, mask: Integer;
  lastC: Integer;
begin
  AddPulse(-665); // direct block mark
  AddPulse(-660); // block data mark

  tps := blk.tPerSample;
  lastC := blk.bitsInLastByte-1; if lastC > 7 then lastC := 7;
  mask := 0; ticks := 0;
  for f := 0 to High(blk.data) do
  begin
    // process byte
    if f = High(blk.data) then cnt := lastC else cnt := 7;
    for c := 0 to cnt do
    begin
      Inc(ticks, tps);
      if ((blk.data[f] xor mask) and ($80 shr c)) <> 0 then
      begin
        // pulse
        AddPulse(ticks);
        mask := not mask; ticks := 0;
      end;
    end;
  end;

  // the last pulse (is this necessary?)
  if ticks > 0 then AddPulse(ticks);
  if blk.pause > 0 then AddPulse(Integer(blk.pause)*3500);
end;

procedure TZXRawPulses.BuildFromPause (blk: TZXBlockPause);
begin
  AddPulse(-667); // pause block mark
  AddPulse(-660); // block data mark
  if blk.length = 0 then
  begin
    // at least 1ms pulse as specified in TZX 1.13
    AddPulse(3500);
  end;
  AddPause(blk.length);
end;


function LoadLunterTapeFromStream (st: TStream): TZXTape;
var
  dsc: TZXBlockDescription;
  std: TZXBlockStandard;
  w: Word;
begin
  result := TZXTape.Create();
  try
    dsc := TZXBlockDescription.Create();
    dsc.text := 'created by DarkSpecter!';
    result.AppendBlock(dsc);

    while st.Position < st.Size do
    begin
      if st.Position+2 > st.Size then
      begin
        WriteLn('invalid .TAP file!');
        break;
      end;
      st.ReadBuffer(w, 2); // length
      if st.Position+w > st.Size then
      begin
        WriteLn('invalid .TAP file (1)!');
        break;
      end;
      std := TZXBlockStandard.Create();
      std.pause := 1000;
      try
        SetLength(std.data, w);
        if w > 0 then st.ReadBuffer(std.data[0], w);
        result.AppendBlock(std);
      except
        std.Free();
        raise;
      end;
    end;
  except
    result.Free();
    raise;
  end;
end;

procedure LunterTapeToTZX (fi, fo: TStream);
const
  TZX_SIGN = 'ZXTape!'#26;
var
  sign: packed array [0..7] of Char;
  ss: ShortString;
  w, sz: Word;
  b: Byte;
begin
  //  
  sign := TZX_SIGN; fo.WriteBuffer(sign[0], 8); // 
  //  (1.10)
  b := 1; fo.WriteBuffer(b, 1);
  b := 10; fo.WriteBuffer(b, 1);
  // 
  ss := 'created by DarkSpecter!';
  b := $30; fo.WriteBuffer(b, 1); //   
  fo.WriteBuffer(ss[0], Length(ss)+1);

  ss := 'empty .TAP block';

  //  
  while fi.Position < fi.Size do
  begin
    if fi.Position+2 > fi.Size then
    begin
      WriteLn('invalid .TAP file!');
      break;
    end;
    fi.ReadBuffer(sz, 2); // length
    if fi.Position+sz > fi.Size then
    begin
      WriteLn('invalid .TAP file (1)!');
      break;
    end;
    if sz = 0 then
    begin
      //  
      b := $30; fo.WriteBuffer(b, 1); //    %-)
      fo.WriteBuffer(ss[0], Length(ss)+1);
    end
    else
    begin
      //  
      b := $10; fi.WriteBuffer(b, 1); // 
      w := 1000; fi.WriteBuffer(w, 2); //   
      fi.WriteBuffer(sz, 2); //  
      if sz > 0 then fo.CopyFrom(fi, sz);
    end;
  end;
end;


end.
